from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class OpenSslConfigPair(Model):
    """
    Key-value pair in the OpenSSL config block

    [ name ]
    key = value
    key2 = value2
    ...

    This model is not expected to be used as a message (produced/consumed by actors).
    See the OpenSslConfig.
    """
    topic = SystemInfoTopic

    key = fields.String()
    """ The key is usually fixed name for specific purpose """
    value = fields.String()
    """ The value, can be a reference to another block """


class OpenSslConfigBlock(Model):
    """
    Every block in the openssl.cnf in the following format:

    [ name ]
    key = value
    key2 = value2
    ...

    This model is not expected to be used as a message (produced/consumed by actors).
    See the OpenSslConfig.
    """
    topic = SystemInfoTopic

    name = fields.String()
    """ The block name """
    pairs = fields.List(fields.Model(OpenSslConfigPair))
    """ The key-value pairs """


class OpenSslConfig(Model):
    """
    openssl.cnf

    This mode contains interesting parts of the RHEL8 OpenSSL configuration file
    that will be later used to decide if it needs to be updated to keep working
    in RHEL9.
    """
    topic = SystemInfoTopic

    openssl_conf = fields.Nullable(fields.String())
    """
    The value of openssl_conf field

    It is used to load default TLS policy in RHEL8, but controls loading of all
    providers in RHEL9 so it needs to be adjusted for upgrade. This is listed
    before any block.
    """

    blocks = fields.List(fields.Model(OpenSslConfigBlock))
    """
    The list of blocks in the openssl.cnf

    We are mostly interested in the ones referenced by the openssl_conf value above.
    """

    modified = fields.Boolean(default=False)
    """ True if the configuration file was modified. """



from leapp.models import (
    DistributionSignedRPM,
    RepositoriesFacts,
    RepositoryData,
    RepositoryFile,
    RPM,
    UsedRepositories
)
from leapp.snactor.fixture import current_actor_context


def get_sample_rpm(name, repository):
    return RPM(
        name=name,
        epoch='1',
        packager='Red Hat Inc.',
        version='0.0.1',
        release='1.el7',
        arch='x86_64',
        pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM UTC, Key ID 0123456789abcdef',
        repository=repository)


def get_sample_installed_pkgs(pkgs):
    return DistributionSignedRPM(items=[get_sample_rpm(*p) for p in pkgs])


def get_sample_repository(repoid, name):
    return RepositoryFile(file='/etc/yum.d/sample.repo', data=[RepositoryData(
        repoid=repoid,
        name=name,
        enabled=True)])


def get_sample_repositories(repos):
    return RepositoriesFacts(
        repositories=[get_sample_repository(*r) for r in repos])


def test_actor_execution(current_actor_context):
    installed = get_sample_installed_pkgs([
        ('pkg1', 'rhel-7-server-rpms'),
        ('pkg2', 'rhel-7-server-rpms')])

    repos = get_sample_repositories([
        ('rhel-7-server-rpms', 'RHEL 7 Server'),
        ('rhel-7-unused-rpms', 'RHEL 7 Unused')])

    current_actor_context.feed(installed)
    current_actor_context.feed(repos)
    current_actor_context.run()
    assert current_actor_context.consume(UsedRepositories)
    used_repos = current_actor_context.consume(UsedRepositories)[0].repositories
    assert len(used_repos) == 1
    assert used_repos[0].repository == 'rhel-7-server-rpms'
    assert len(used_repos[0].packages) == 2
    assert used_repos[0].packages == ['pkg1', 'pkg2']



from leapp.libraries.actor import repairsystemdsymlinks
from leapp.libraries.common import systemd
from leapp.libraries.common.testutils import CurrentActorMocked, logger_mocked
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import (
    SystemdBrokenSymlinksSource,
    SystemdBrokenSymlinksTarget,
    SystemdServiceFile,
    SystemdServicesInfoSource
)


class MockedSystemdCmd(object):
    def __init__(self):
        self.units = []

    def __call__(self, unit, *args, **kwargs):
        self.units.append(unit)
        return {}


def test_bad_symslinks(monkeypatch):
    service_files = [
        SystemdServiceFile(name='rngd.service', state='enabled'),
        SystemdServiceFile(name='sysstat.service', state='disabled'),
        SystemdServiceFile(name='hello.service', state='enabled'),
        SystemdServiceFile(name='world.service', state='disabled'),
    ]

    def is_unit_enabled_mocked(unit):
        return True

    monkeypatch.setattr(repairsystemdsymlinks, '_is_unit_enabled', is_unit_enabled_mocked)

    reenable_mocked = MockedSystemdCmd()
    monkeypatch.setattr(systemd, 'reenable_unit', reenable_mocked)

    service_info = SystemdServicesInfoSource(service_files=service_files)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[service_info]))

    repairsystemdsymlinks._handle_bad_symlinks(service_info.service_files)

    assert reenable_mocked.units == ['rngd.service']


def test_handle_newly_broken_symlink(monkeypatch):

    symlinks = [
        '/etc/systemd/system/default.target.wants/systemd-readahead-replay.service',
        '/etc/systemd/system/multi-user.target.wants/vdo.service',
        '/etc/systemd/system/multi-user.target.wants/hello.service',
        '/etc/systemd/system/multi-user.target.wants/world.service',
        '/etc/systemd/system/multi-user.target.wants/foo.service',
        '/etc/systemd/system/multi-user.target.wants/bar.service',
    ]

    def is_unit_enabled_mocked(unit):
        return unit in ('hello.service', 'foo.service')

    expect_disabled = [
        'systemd-readahead-replay.service',
        'vdo.service',
        'world.service',
        'bar.service',
    ]

    expect_reenabled = [
        'hello.service',
    ]

    monkeypatch.setattr(repairsystemdsymlinks, '_is_unit_enabled', is_unit_enabled_mocked)

    reenable_mocked = MockedSystemdCmd()
    monkeypatch.setattr(systemd, 'reenable_unit', reenable_mocked)

    disable_mocked = MockedSystemdCmd()
    monkeypatch.setattr(systemd, 'disable_unit', disable_mocked)

    service_files = [
        SystemdServiceFile(name='systemd-readahead-replay.service', state='enabled'),
        SystemdServiceFile(name='vdo.service', state='disabled'),
        SystemdServiceFile(name='hello.service', state='enabled'),
        SystemdServiceFile(name='world.service', state='disabled'),
        SystemdServiceFile(name='foo.service', state='disabled'),
        SystemdServiceFile(name='bar.service', state='enabled'),
    ]
    service_info = SystemdServicesInfoSource(service_files=service_files)
    repairsystemdsymlinks._handle_newly_broken_symlinks(symlinks, service_info)

    assert reenable_mocked.units == expect_reenabled
    assert disable_mocked.units == expect_disabled



from leapp.actors import Actor
from leapp.libraries.actor.removebootfiles import remove_boot_files
from leapp.models import BootContent
from leapp.tags import IPUWorkflowTag, PreparationPhaseTag


class RemoveBootFiles(Actor):
    """
    Remove Leapp provided initramfs from boot partition.

    Since Leapp provided initramfs and kernel are already loaded into RAM at this phase, remove
    them to have as little space requirements for boot partition as possible.
    """

    name = 'remove_boot_files'
    consumes = (BootContent,)
    produces = ()
    tags = (IPUWorkflowTag, PreparationPhaseTag)

    def process(self):
        remove_boot_files()



import glob
import os

import pytest

from leapp import reporting
from leapp.libraries.actor import scandynamiclinkerconfiguration
from leapp.libraries.common.testutils import produce_mocked
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.models import DistributionSignedRPM

INCLUDED_CONFIGS_GLOB_DICT_1 = {'/etc/ld.so.conf.d/*.conf': ['/etc/ld.so.conf.d/dyninst-x86_64.conf',
                                                             '/etc/ld.so.conf.d/mariadb-x86_64.conf',
                                                             '/etc/ld.so.conf.d/bind-export-x86_64.conf']}

INCLUDED_CONFIGS_GLOB_DICT_2 = {'/etc/ld.so.conf.d/*.conf': ['/etc/ld.so.conf.d/dyninst-x86_64.conf',
                                                             '/etc/ld.so.conf.d/mariadb-x86_64.conf',
                                                             '/etc/ld.so.conf.d/bind-export-x86_64.conf',
                                                             '/etc/ld.so.conf.d/custom1.conf',
                                                             '/etc/ld.so.conf.d/custom2.conf']}

INCLUDED_CONFIGS_GLOB_DICT_3 = {'/etc/ld.so.conf.d/*.conf': ['/etc/ld.so.conf.d/dyninst-x86_64.conf',
                                                             '/etc/ld.so.conf.d/custom1.conf',
                                                             '/etc/ld.so.conf.d/mariadb-x86_64.conf',
                                                             '/etc/ld.so.conf.d/bind-export-x86_64.conf',
                                                             '/etc/ld.so.conf.d/custom2.conf'],
                                '/custom/path/*.conf': ['/custom/path/custom1.conf',
                                                        '/custom/path/custom2.conf']}


@pytest.mark.parametrize(('included_configs_glob_dict', 'other_lines', 'custom_configs', 'used_variables'),
                         [
                            (INCLUDED_CONFIGS_GLOB_DICT_1, [], [], []),
                            (INCLUDED_CONFIGS_GLOB_DICT_1, ['/custom/path.lib'], [], []),
                            (INCLUDED_CONFIGS_GLOB_DICT_1, [], [], ['LD_LIBRARY_PATH']),
                            (INCLUDED_CONFIGS_GLOB_DICT_2, [], ['/etc/ld.so.conf.d/custom1.conf',
                                                                '/etc/ld.so.conf.d/custom2.conf'], []),
                            (INCLUDED_CONFIGS_GLOB_DICT_3, ['/custom/path.lib'], ['/etc/ld.so.conf.d/custom1.conf',
                                                                                  '/etc/ld.so.conf.d/custom2.conf'
                                                                                  '/custom/path/custom1.conf',
                                                                                  '/custom/path/custom2.conf'], []),
                        ])
def test_scan_dynamic_linker_configuration(monkeypatch, included_configs_glob_dict, other_lines,
                                           custom_configs, used_variables):
    monkeypatch.setattr(scandynamiclinkerconfiguration, '_parse_main_config',
                        lambda: (included_configs_glob_dict.keys(), other_lines))
    monkeypatch.setattr(glob, 'glob', lambda glob: included_configs_glob_dict[glob])
    monkeypatch.setattr(scandynamiclinkerconfiguration, '_is_included_config_custom',
                        lambda config: config in custom_configs)
    monkeypatch.setattr(api, 'produce', produce_mocked())

    for var in used_variables:
        monkeypatch.setenv(var, '/some/path')

    scandynamiclinkerconfiguration.scan_dynamic_linker_configuration()

    produce_expected = custom_configs or other_lines or used_variables
    if not produce_expected:
        assert not api.produce.called
        return

    assert api.produce.called == 1

    configuration = api.produce.model_instances[0]

    all_configs = []
    for configs in included_configs_glob_dict.values():
        all_configs += configs

    assert len(all_configs) == len(configuration.included_configs)
    for config in configuration.included_configs:
        if config.path in custom_configs:
            assert config.modified

    assert configuration.main_config.path == scandynamiclinkerconfiguration.LD_SO_CONF_MAIN
    if other_lines:
        assert configuration.main_config.modified
        assert configuration.main_config.modified_lines == other_lines

    if used_variables:
        assert configuration.used_variables == used_variables


@pytest.mark.parametrize(('config_contents', 'included_config_paths', 'other_lines'),
                         [
                            (['include ld.so.conf.d/*.conf\n'],
                             ['/etc/ld.so.conf.d/*.conf'], []),
                            (['include ld.so.conf.d/*.conf\n', '\n', '/custom/path.lib\n', '#comment'],
                             ['/etc/ld.so.conf.d/*.conf'], ['/custom/path.lib']),
                            (['include ld.so.conf.d/*.conf\n', 'include /custom/path.conf\n'],
                             ['/etc/ld.so.conf.d/*.conf', '/custom/path.conf'], []),
                            (['include ld.so.conf.d/*.conf\n', '#include /custom/path.conf\n', '#/custom/path.conf\n'],
                             ['/etc/ld.so.conf.d/*.conf'], []),
                            ([' \n'],
                             [], [])
                         ])
def test_parse_main_config(monkeypatch, config_contents, included_config_paths, other_lines):
    def mocked_read_file(path):
        assert path == scandynamiclinkerconfiguration.LD_SO_CONF_MAIN
        return config_contents

    monkeypatch.setattr(scandynamiclinkerconfiguration, '_read_file', mocked_read_file)

    _included_config_paths, _other_lines = scandynamiclinkerconfiguration._parse_main_config()

    assert _included_config_paths == included_config_paths
    assert _other_lines == other_lines


@pytest.mark.parametrize(('config_path', 'run_result', 'is_modified'),
                         [
                            ('/etc/ld.so.conf.d/dyninst-x86_64.conf',
                             '.......T.  c /etc/ld.so.conf.d/dyninst-x86_64.conf', False),
                            ('/etc/ld.so.conf.d/dyninst-x86_64.conf',
                             'S.5....T.  c /etc/ld.so.conf.d/dyninst-x86_64.conf', True),
                            ('/etc/ld.so.conf.d/kernel-3.10.0-1160.el7.x86_64.conf',
                             '', False)
                         ])
def test_is_modified(monkeypatch, config_path, run_result, is_modified):
    def mocked_run(command, checked):
        assert config_path in command
        assert checked is False
        exit_code = 1 if run_result else 0
        return {'stdout': run_result, 'exit_code': exit_code}

    monkeypatch.setattr(scandynamiclinkerconfiguration, 'run', mocked_run)

    _is_modified = scandynamiclinkerconfiguration._is_modified(config_path)
    assert _is_modified == is_modified


@pytest.mark.parametrize(('config_path',
                          'config_contents', 'run_result',
                          'is_installed_rh_signed_package', 'is_modified', 'has_effective_lines'),
                         [
                            ('/etc/ld.so.conf.d/dyninst-x86_64.conf',
                             ['/usr/lib64/dyninst\n'], 'dyninst',
                             True, False, True),  # RH sighend package without modification - Not custom
                            ('/etc/ld.so.conf.d/dyninst-x86_64.conf',
                             ['/usr/lib64/my_dyninst\n'], 'dyninst',
                             True, True, True),  # Was modified by user - Custom
                            ('/etc/custom/custom.conf',
                             ['/usr/lib64/custom'], 'custom',
                             False, None, True),  # Third-party package - Custom
                            ('/etc/custom/custom.conf',
                             ['#/usr/lib64/custom\n'], 'custom',
                             False, None, False),  # Third-party package without effective lines - Not custom
                            ('/etc/ld.so.conf.d/somelib.conf',
                             ['/usr/lib64/somelib\n'], CalledProcessError,
                             None, None, True),  # User created configuration file - Custom
                            ('/etc/ld.so.conf.d/somelib.conf',
                             ['#/usr/lib64/somelib\n'], CalledProcessError,
                             None, None, False)  # User created configuration file without effective lines - Not custom
                         ])
def test_is_included_config_custom(monkeypatch, config_path, config_contents, run_result,
                                   is_installed_rh_signed_package, is_modified, has_effective_lines):
    def mocked_run(command):
        assert config_path in command
        if run_result and not isinstance(run_result, str):
            raise CalledProcessError("message", command, "result")
        return {'stdout': run_result}

    def mocked_has_package(model, package_name):
        assert model is DistributionSignedRPM
        assert package_name == run_result
        return is_installed_rh_signed_package

    def mocked_read_file(path):
        assert path == config_path
        return config_contents

    monkeypatch.setattr(scandynamiclinkerconfiguration, 'run', mocked_run)
    monkeypatch.setattr(scandynamiclinkerconfiguration, 'has_package', mocked_has_package)
    monkeypatch.setattr(scandynamiclinkerconfiguration, '_read_file', mocked_read_file)
    monkeypatch.setattr(scandynamiclinkerconfiguration, '_is_modified', lambda *_: is_modified)
    monkeypatch.setattr(os.path, 'isfile', lambda _: True)

    result = scandynamiclinkerconfiguration._is_included_config_custom(config_path)
    is_custom = not isinstance(run_result, str) or not is_installed_rh_signed_package or is_modified
    is_custom &= has_effective_lines
    assert result == is_custom



from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM


def _append_string(path, content):
    """
    Appends string into file

    :param str path: path to file
    :param str content: content to add
    """
    with open(path, 'a') as f:
        f.write(content)


# rpm : the default config file
vim_configs = {
    'vim-minimal': '/etc/virc',
    'vim-enhanced': '/etc/vimrc'
}


# list of macros that should be set
new_macros = [
    'let skip_defaults_vim=1',
    'set t_BE='
]


def update_config(path, append_function=_append_string):
    """
    Insert expected content into the file on the path

    :param str path: string representing the full path of the config file
    :param func append_function: appends string into file
    """
    fmt_input = "\n{comment_line}\n{content}\n".format(comment_line='" content added by Leapp',
                                                       content='\n'.join(new_macros))

    try:
        append_function(path, fmt_input)
    except IOError:
        raise IOError('Error during writing to file: {}.'.format(path))


def _check_package(pkg):
    """
    Checks if a package is installed and signed

    :param str pkg: name of package
    """
    return has_package(DistributionSignedRPM, pkg)


def update_vim(debug_log=api.current_logger().debug,
               error_log=api.current_logger().error,
               is_installed=_check_package,
               append_function=_append_string):
    """
    Do update of configuration files

    :param func debug_log: function for debug logging
    :param func error_log: function for error logging
    :param func is_installed: checks if a package is installed
    :param func append_function: appends string into file
    """
    error_list = []

    for pkg, config_file in vim_configs.items():
        if not is_installed(pkg):
            continue

        debug_log('Updating Vim configuration file {}.'.format(config_file))

        try:
            update_config(config_file, append_function)
        except (OSError, IOError) as error:
            error_list.append((config_file, error))
    if error_list:
        error_log('The files below have not been modified (error message included):' +
                  ''.join(['\n    - {}: {}'.format(err[0], err[1]) for err in error_list]))



from leapp import reporting
from leapp.libraries.stdlib import api
from leapp.models import DynamicLinkerConfiguration

LD_SO_CONF_DIR = '/etc/ld.so.conf.d'
LD_SO_CONF_MAIN = '/etc/ld.so.conf'
LD_LIBRARY_PATH_VAR = 'LD_LIBRARY_PATH'
LD_PRELOAD_VAR = 'LD_PRELOAD'
FMT_LIST_SEPARATOR_1 = '\n- '
FMT_LIST_SEPARATOR_2 = '\n    - '


def _report_custom_dynamic_linker_configuration(summary):
    reporting.create_report([
        reporting.Title(
            'Detected customized configuration for dynamic linker.'
        ),
        reporting.Summary(summary),
        reporting.Remediation(hint=('Remove or revert the custom dynamic linker configurations and apply the changes '
                                    'using the ldconfig command. In case of possible active software collections we '
                                    'suggest disabling them persistently.')),
        reporting.RelatedResource('file', '/etc/ld.so.conf'),
        reporting.RelatedResource('directory', '/etc/ld.so.conf.d'),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.OS_FACTS]),
    ])


def check_dynamic_linker_configuration():
    configuration = next(api.consume(DynamicLinkerConfiguration), None)
    if not configuration:
        return

    custom_configurations = ''
    if configuration.main_config.modified:
        custom_configurations += (
            '{}The {} file has unexpected contents:{}{}'
            .format(FMT_LIST_SEPARATOR_1, LD_SO_CONF_MAIN,
                    FMT_LIST_SEPARATOR_2, FMT_LIST_SEPARATOR_2.join(configuration.main_config.modified_lines))
        )

    custom_configs = []
    for config in configuration.included_configs:
        if config.modified:
            custom_configs.append(config.path)

    if custom_configs:
        custom_configurations += (
            '{}The following drop in config files were marked as custom:{}{}'
            .format(FMT_LIST_SEPARATOR_1, FMT_LIST_SEPARATOR_2, FMT_LIST_SEPARATOR_2.join(custom_configs))
        )

    if configuration.used_variables:
        custom_configurations += (
            '{}The following variables contain unexpected dynamic linker configuration:{}{}'
            .format(FMT_LIST_SEPARATOR_1, FMT_LIST_SEPARATOR_2,
                    FMT_LIST_SEPARATOR_2.join(configuration.used_variables))
        )

    if custom_configurations:
        summary = (
            'Custom configurations to the dynamic linker could potentially impact '
            'the upgrade in a negative way. The custom configuration includes '
            'modifications to {main_conf}, custom or modified drop in config '
            'files in the {conf_dir} directory and additional entries in the '
            '{ldlib_envar} or {ldpre_envar} variables. These modifications '
            'configure the dynamic linker to use different libraries that might '
            'not be provided by Red Hat products or might not be present during '
            'the whole upgrade process. The following custom configurations '
            'were detected by leapp:{cust_configs}'
            .format(
                main_conf=LD_SO_CONF_MAIN,
                conf_dir=LD_SO_CONF_DIR,
                ldlib_envar=LD_LIBRARY_PATH_VAR,
                ldpre_envar=LD_PRELOAD_VAR,
                cust_configs=custom_configurations
            )
        )
        _report_custom_dynamic_linker_configuration(summary)



import os

from leapp.actors import Actor
from leapp.libraries.common.config import architecture
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import run
from leapp.models import (
    DNFWorkaround,
    InstalledRPM,
    Module,
    RepositoriesSetupTasks,
    RpmTransactionTasks,
    SatelliteFacts,
    SatellitePostgresqlFacts
)
from leapp.tags import FactsPhaseTag, IPUWorkflowTag

POSTGRESQL_SCL_DATA_PATH = '/var/opt/rh/rh-postgresql12/lib/pgsql/data/'


class SatelliteUpgradeFacts(Actor):
    """
    Report which Satellite packages require updates and how to handle PostgreSQL data
    """

    name = 'satellite_upgrade_facts'
    consumes = (InstalledRPM, )
    produces = (DNFWorkaround, RepositoriesSetupTasks, RpmTransactionTasks, SatelliteFacts)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        if not architecture.matches_architecture(architecture.ARCH_X86_64):
            return

        has_foreman = has_package(InstalledRPM, 'foreman') or has_package(InstalledRPM, 'foreman-proxy')
        if not has_foreman:
            return

        has_katello_installer = has_package(InstalledRPM, 'foreman-installer-katello')

        local_postgresql = has_package(InstalledRPM, 'rh-postgresql12-postgresql-server')
        postgresql_contrib = has_package(InstalledRPM, 'rh-postgresql12-postgresql-contrib')
        postgresql_evr = has_package(InstalledRPM, 'rh-postgresql12-postgresql-evr')

        # SCL-related packages
        to_remove = ['tfm-runtime', 'tfm-pulpcore-runtime', 'rh-redis5-runtime', 'rh-ruby27-runtime',
                     'rh-python38-runtime']
        to_install = ['rubygem-foreman_maintain']
        modules_to_enable = [Module(name='ruby', stream='2.7')]

        if has_package(InstalledRPM, 'katello'):
            # enable modules that are needed for Candlepin, which is pulled in by Katello
            modules_to_enable.append(Module(name='pki-core', stream='10.6'))
            modules_to_enable.append(Module(name='pki-deps', stream='10.6'))
            # enable modules that are needed for Pulpcore
            modules_to_enable.append(Module(name='python38', stream='3.8'))
            to_install.append('katello')
            # Force removal of tomcat
            # PES data indicates tomcat.el7 can be upgraded to tomcat.el8 since EL 8.8,
            # but we need pki-servlet-engine from the module instead which will be pulled in via normal
            # package dependencies
            to_remove.extend(['tomcat', 'tomcat-lib'])

        if has_package(InstalledRPM, 'rh-redis5-redis'):
            modules_to_enable.append(Module(name='redis', stream='5'))
            to_install.append('redis')

        for rpm_pkgs in self.consume(InstalledRPM):
            for pkg in rpm_pkgs.items:
                if (pkg.name.startswith('tfm-rubygem-hammer') or pkg.name.startswith('tfm-rubygem-foreman')
                        or pkg.name.startswith('tfm-rubygem-katello')
                        or pkg.name.startswith('tfm-rubygem-smart_proxy')):
                    to_install.append(pkg.name.replace('tfm-rubygem-', 'rubygem-'))
                elif pkg.name.startswith('tfm-pulpcore-python3-pulp'):
                    to_install.append(pkg.name.replace('tfm-pulpcore-python3-', 'python38-'))
                elif pkg.name.startswith('foreman-installer') or pkg.name.startswith('satellite-installer'):
                    to_install.append(pkg.name)

        on_same_partition = True
        bytes_required = None
        bytes_available = None
        old_pgsql_data = False

        if local_postgresql:
            """
            Handle migration of the PostgreSQL legacy-actions files.
            RPM cannot handle replacement of directories by symlinks by default
            without the %pretrans scriptlet. As PostgreSQL package is packaged wrong,
            we have to workaround that by migration of the PostgreSQL files
            before the rpm transaction is processed.
            """
            self.produce(
                DNFWorkaround(
                    display_name='PostgreSQL symlink fix',
                    script_path=self.get_tool_path('handle-postgresql-legacy-actions'),
                )
            )

            old_pgsql_data = bool(os.path.exists('/var/lib/pgsql/data/') and os.listdir('/var/lib/pgsql/data/')
                                  and os.path.exists(POSTGRESQL_SCL_DATA_PATH)
                                  and os.listdir(POSTGRESQL_SCL_DATA_PATH))
            scl_psql_stat = os.stat(POSTGRESQL_SCL_DATA_PATH)
            for nonscl_path in ['/var/lib/pgsql/data/', '/var/lib/pgsql/', '/var/lib/', '/']:
                if os.path.exists(nonscl_path):
                    nonscl_psql_stat = os.stat(nonscl_path)
                    break

            if scl_psql_stat.st_dev != nonscl_psql_stat.st_dev:
                on_same_partition = False
                # get the current disk usage of the PostgreSQL data
                scl_du_call = run(['du', '--block-size=1', '--summarize', POSTGRESQL_SCL_DATA_PATH])
                bytes_required = int(scl_du_call['stdout'].split()[0])
                # get the current free space on the target partition
                nonscl_stat = os.statvfs(nonscl_path)
                bytes_available = nonscl_stat.f_bavail * nonscl_stat.f_frsize

            modules_to_enable.append(Module(name='postgresql', stream='12'))
            to_remove.append('rh-postgresql12-runtime')
            to_install.extend(['postgresql', 'postgresql-server'])
            if postgresql_contrib:
                to_remove.append('rh-postgresql12-postgresql-contrib')
                to_install.append('postgresql-contrib')
            if postgresql_evr:
                to_remove.append('rh-postgresql12-postgresql-evr')
                to_install.append('postgresql-evr')

        self.produce(SatelliteFacts(
            has_foreman=has_foreman,
            has_katello_installer=has_katello_installer,
            postgresql=SatellitePostgresqlFacts(
                local_postgresql=local_postgresql,
                old_var_lib_pgsql_data=old_pgsql_data,
                same_partition=on_same_partition,
                space_required=bytes_required,
                space_available=bytes_available,
            ),
        ))

        repositories_to_enable = ['satellite-maintenance-6.11-for-rhel-8-x86_64-rpms']
        if has_package(InstalledRPM, 'satellite'):
            repositories_to_enable.append('satellite-6.11-for-rhel-8-x86_64-rpms')
            modules_to_enable.append(Module(name='satellite', stream='el8'))
            to_install.append('satellite')
        elif has_package(InstalledRPM, 'satellite-capsule'):
            repositories_to_enable.append('satellite-capsule-6.11-for-rhel-8-x86_64-rpms')
            modules_to_enable.append(Module(name='satellite-capsule', stream='el8'))
            to_install.append('satellite-capsule')

        self.produce(RpmTransactionTasks(
            to_remove=to_remove,
            to_install=to_install,
            modules_to_enable=modules_to_enable
            )
        )

        self.produce(RepositoriesSetupTasks(to_enable=repositories_to_enable))



from leapp.libraries.common import multipathutil

_regexes = ('vendor', 'product', 'revision', 'product_blacklist', 'devnode',
            'wwid', 'property', 'protocol')


def _update_config(need_foreign, need_allow_usb, config):
    if not (need_foreign or need_allow_usb or config.invalid_regexes_exist):
        return None
    contents = multipathutil.read_config(config.pathname)
    if contents is None:
        return None
    lines = contents.split('\n')

    section = None
    in_subsection = False
    updated_file = False
    defaults_start = -1
    for i, line in enumerate(lines):
        try:
            data = multipathutil.LineData(line, section, in_subsection)
        except ValueError:
            continue
        if data.type == data.TYPE_SECTION_END:
            if in_subsection:
                in_subsection = False
            elif section is not None:
                section = None
        elif data.type == data.TYPE_SECTION_START:
            if section is None:
                section = data.section
                if section == 'defaults':
                    defaults_start = i + 1
            elif not in_subsection:
                in_subsection = True
        elif data.type == data.TYPE_OPTION:
            if section == 'defaults':
                if data.option == 'enable_foreign':
                    need_foreign = False
                elif data.option == 'allow_usb_devices':
                    need_allow_usb = False
            if data.option in _regexes and data.value == '*':
                lines[i] = line.replace('*', '.*', 1)
                lines[i] += ' # line modified by Leapp'
                updated_file = True

    if need_foreign or need_allow_usb:
        updated_file = True
        if defaults_start < 0:
            if in_subsection:
                lines.append('\t} # line added by Leapp')
            if section is not None:
                lines.append('} # line added by Leapp')
            lines.append('defaults { # section added by Leapp')
            if need_foreign:
                lines.append('\tenable_foreign ""')
            if need_allow_usb:
                lines.append('\tallow_usb_devices yes')
            lines.append('}')
            lines.append('')
        else:
            if need_allow_usb:
                lines.insert(defaults_start, '\tallow_usb_devices yes # line added by Leapp')
            if need_foreign:
                lines.insert(defaults_start, '\tenable_foreign "" # line added by Leapp')

    if not updated_file:
        return None

    contents = '\n'.join(lines)
    return contents


def update_configs(facts):
    need_foreign = not any(x for x in facts.configs if x.enable_foreign_exists)
    need_allow_usb = not any(x for x in facts.configs if x.allow_usb_exists)
    for config in facts.configs:
        contents = _update_config(need_foreign, need_allow_usb, config)
        need_foreign = False
        need_allow_usb = False
        """
        foreign_exists and allow_usb_exists only matter for the main
        config file.
        """
        if contents:
            multipathutil.write_config(config.pathname, contents)



from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api
from leapp.models import CopyFile, TargetUserSpacePreupgradeTasks


def _get_files_to_copy(cpi):
    return [f.path for f in cpi.custom_policies + cpi.custom_modules]


def process(cpi_messages):
    cpi = next(cpi_messages, None)
    if list(cpi_messages):
        api.current_logger().warning('Unexpectedly received more than one CryptoPolicyInfo message.')
    if not cpi:
        raise StopActorExecutionError(
            'Could not check crypto policies status',
            details={'details': 'No CryptoPolicyInfo facts found.'}
        )

    if cpi.current_policy != 'DEFAULT':
        # If we have to change the crypto policies inside the target userspace container,
        # we need update-crypto-policies script inside as well as potential custom policies files
        files = [CopyFile(src=f) for f in _get_files_to_copy(cpi)]
        api.produce(TargetUserSpacePreupgradeTasks(install_rpms=['crypto-policies-scripts'],
                                                   copy_files=files))

        # When non-default crypto policy is used, it might be outdated. Recommend user to revisit.
        # exceptions are here the FIPS and FUTURE policies, which are more future-proof.
        if cpi.current_policy in ('FIPS', 'FUTURE'):
            return
        reporting.create_report([
            reporting.Title('System-wide crypto policy is set to non-DEFAULT policy'),
            reporting.Summary((
                    "The system-wide crypto policies are set to `{}` value. This might be"
                    " outdated decision, the custom crypto policy might be outdated and no"
                    " longer meeting the security standards. Please, review the current crypto"
                    " policies settings. If this is intentional and up-to-date, you can ignore"
                    " this message. The custom crypto policy will be configured on the updated"
                    " system."
                ).format(cpi.current_policy)),
            reporting.Severity(reporting.Severity.MEDIUM),
            reporting.Groups([reporting.Groups.SECURITY, reporting.Groups.SANITY]),
            reporting.Remediation(hint="Review the current policy in /etc/crypto-policies/state/CURRENT.pol"),
            reporting.RelatedResource('package', 'crypto-policies'),
            reporting.ExternalLink(
                url='https://red.ht/rhel-9-security-considerations',
                title='Security Considerations in adopting RHEL 9'
            )
        ])



import pytest

from leapp.libraries.actor import grubenvtofile
from leapp.libraries.common.testutils import logger_mocked
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.models import HybridImage


def raise_call_error(args=None):
    raise CalledProcessError(
        message='A Leapp Command Error occurred.',
        command=args,
        result={'signal': None, 'exit_code': 1, 'pid': 0, 'stdout': 'fake', 'stderr': 'fake'}
    )


class run_mocked(object):
    def __init__(self, raise_err=False):
        self.called = 0
        self.args = []
        self.raise_err = raise_err

    def __call__(self, *args):
        self.called += 1
        self.args.append(args)
        if self.raise_err:
            raise_call_error(args)


def test_grubenv_to_file(monkeypatch):
    monkeypatch.setattr(api, 'consume', lambda x: iter([HybridImage()]))
    monkeypatch.setattr(grubenvtofile, 'run', run_mocked())
    grubenvtofile.grubenv_to_file()
    assert grubenvtofile.run.called == 2


def test_fail_grubenv_to_file(monkeypatch):
    monkeypatch.setattr(api, 'consume', lambda x: iter([HybridImage()]))
    monkeypatch.setattr(grubenvtofile, 'run', run_mocked(raise_err=True))
    monkeypatch.setattr(api, 'current_logger', logger_mocked())
    grubenvtofile.grubenv_to_file()
    assert grubenvtofile.run.called == 1
    assert api.current_logger.warnmsg[0].startswith('Could not unlink')



import pytest
import six

from leapp.models import InvalidRootSubdirectory, Report, RootDirectory, RootSubdirectory
from leapp.snactor.fixture import current_actor_context


@pytest.mark.parametrize('modified,invalid,expected_report', [
    ({}, {}, False),
    ({'bin': '/usr/bin'}, {}, True),
    ({'sbin': 'usr/bin'}, {}, False),
    ({'sbin': '/usr/sbin'}, {}, True),
    ({'lib': '/usr/lib', 'lib64': 'usr/lib64'}, {}, True),
    ({}, {b'\xc4\xc9\xd2\xc5\xcb\xd4\xcf\xd2\xc9\xd1': b''}, False),
    ({}, {b'\xc4\xc9\xd2\xc5\xcb\xd4\xcf\xd2\xc9\xd1': b'usr/bin'}, False),
    ({'lib': '/usr/lib', 'lib64': 'usr/lib64'}, {b'\xc4\xc9\xd2\xc5\xcb\xd4\xcf\xd2\xc9\xd1': b'/usr/lib64'}, True),
    ({}, {b'\xc4\xc9\xd2\xc5\xcb\xd4\xcf\xd2\xc9\xd1': b'/usr/lib64'}, True)])
def test_wrong_symlink_inhibitor(current_actor_context, modified, invalid, expected_report):
    invalid_subdirs = {}
    subdirs = {
        'bin': 'usr/bin',
        'boot': '',
        'dev': '',
        'etc': '',
        'home': '',
        'lib': 'usr/lib',
        'lib64': 'usr/lib64',
        'media': '',
        'mnt': '',
        'opt': '',
        'proc': '',
        'root': '',
        'run': '',
        'sbin': 'usr/sbin',
        'srv': '',
        'sys': '',
        'tmp': '',
        'usr': '',
        'var': ''
    }
    subdirs.update(modified)
    invalid_subdirs.update(invalid)

    items = [RootSubdirectory(name=name, target=target) for name, target in subdirs.items()]
    invalid_items = [InvalidRootSubdirectory(name=name, target=target) for name, target in invalid_subdirs.items()]
    current_actor_context.feed(RootDirectory(items=items, invalid_items=invalid_items))
    current_actor_context.run()
    if expected_report:
        report = current_actor_context.consume(Report)
        assert report
        # Make sure the hint is there in case of non-utf bad symlinks
        if invalid_subdirs:
            msg = 'symbolic links point to absolute paths that have non-utf8 encoding'
            hint = next((rem['context'] for rem in report[0].report['detail']['remediations']
                         if rem['type'] == 'hint'), None)
            assert msg in hint
    else:
        assert not current_actor_context.consume(Report)



from leapp.actors import Actor
from leapp.libraries.actor.nischeck import report_nis
from leapp.models import DistributionSignedRPM, NISConfig, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class NISCheck(Actor):
    """
    Checks if any of NIS components is installed and configured
    on the system and warns users about discontinuation.
    """

    name = 'nis_check'
    consumes = (DistributionSignedRPM, NISConfig)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        report_nis()



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class CustomCryptoPolicy(Model):
    """
    Provides information about custom crypto policy found on the source system.
    """
    topic = SystemInfoTopic

    name = fields.String()
    """
    The policy name, derived from the filename.
    """

    path = fields.String()
    """
    The path to the policy file to be copied to the target system.
    """


class CustomCryptoPolicyModule(CustomCryptoPolicy):
    """
    Internally, this carries the same information as CustomCryptoPolicy model,
    but the path will point to different directory as the semantics of the files
    is different.
    """


class CryptoPolicyInfo(Model):
    """
    Provide information related to crypto policies
    """
    topic = SystemInfoTopic

    current_policy = fields.String()
    """
    The current used crypto policy: /etc/crypto-policies/state/current

    Contains e.g. 'LEGACY', 'DEFAULT', ...
    """

    custom_policies = fields.List(fields.Model(CustomCryptoPolicy))
    """
    This is the list of custom crypto policies with *.pol extension found under the following
    directories that are not part of any RPM package:

    * /usr/share/crypto-policies/policies/
    * /etc/crypto-policies/policies/

    """

    custom_modules = fields.List(fields.Model(CustomCryptoPolicyModule))
    """
    This is the list of custom crypto policies modules with *.pmod extension found under the
    following directories that are not part of any RPM package:

    * /usr/share/crypto-policies/policies/modules/
    * /etc/crypto-policies/policies/modules/

    """



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class PamService(Model):
    """
    Pam service description

    This model contains information about pam modules used by specific PAM
    service/filename
    """
    topic = SystemInfoTopic

    service = fields.String()
    modules = fields.List(fields.String())
    # Should this also list includes?


class PamConfiguration(Model):
    """
    Global PAM configuration

    This model describes separate services using PAM and what pam modules are
    used in each of them. Consumer can select just the pam services he is
    interested in or scan for specific configuration throughout all the services.
    """
    topic = SystemInfoTopic

    services = fields.List(fields.Model(PamService))



from leapp.libraries.actor import spamassassinconfigread_spamc, spamassassinconfigread_spamd
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM, SpamassassinFacts


def is_processable():
    """
    Checks whether the spamassassin package is installed.
    """
    res = has_package(DistributionSignedRPM, 'spamassassin')
    if not res:
        api.current_logger().debug('spamassassin is not installed.')
    return res


def get_spamassassin_facts(read_func, listdir):
    """
    Reads the spamc configuration file, the spamassassin sysconfig file and checks
    whether the spamassassin service is overridden. Returns SpamassassinFacts.
    """
    spamc_ssl_argument = spamassassinconfigread_spamc.get_spamc_ssl_argument(read_func)
    service_overriden = spamassassinconfigread_spamd.spamassassin_service_overriden(listdir)
    spamd_ssl_version = spamassassinconfigread_spamd.get_spamd_ssl_version(read_func)
    return SpamassassinFacts(spamc_ssl_argument=spamc_ssl_argument,
                             service_overriden=service_overriden,
                             spamd_ssl_version=spamd_ssl_version)



from leapp.actors import Actor
from leapp.models import DistributionSignedRPM, RepositoriesFacts, UsedRepositories, UsedRepository
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class UsedRepositoriesScanner(Actor):
    """
    Scan used enabled repositories

    Based on lists of installed RPM packages and enabled RPM repositories, check which packages
    were installed from each repository.
    """

    name = 'used_repository_scanner'
    consumes = (DistributionSignedRPM, RepositoriesFacts)
    produces = (UsedRepositories,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        enabled_repos = []
        for repos in self.consume(RepositoriesFacts):
            for repo_file in repos.repositories:
                for repo in repo_file.data:
                    if repo.enabled:
                        enabled_repos.append(repo.repoid)

        installed_pkgs = []
        for rpm_pkgs in self.consume(DistributionSignedRPM):
            installed_pkgs.extend(rpm_pkgs.items)

        used_repos = {}
        for pkg in installed_pkgs:
            if pkg.repository in enabled_repos:
                used_repos.setdefault(pkg.repository, []).append(pkg.name)

        result = UsedRepositories()
        for repo, pkgs in used_repos.items():
            result.repositories.append(UsedRepository(repository=repo, packages=pkgs))
        self.produce(result)



import json
import logging
import os
from collections import namedtuple

from leapp import reporting
from leapp.libraries.common.config import architecture
from leapp.models import EnvVar
from leapp.utils.deprecation import deprecated


class produce_mocked(object):
    def __init__(self):
        self.called = 0
        self.model_instances = []

    def __call__(self, *model_instances):
        self.called += len(model_instances)
        self.model_instances.extend(list(model_instances))


class create_report_mocked(object):
    def __init__(self):
        self.called = 0
        self.reports = []

    def __call__(self, report_fields):
        self.called += 1
        report_obj = reporting._create_report_object(report_fields)
        full_report = json.loads(report_obj.dump()['report'])
        self.reports.append(full_report)

    @property
    def report_fields(self):
        if self.reports:
            return self.reports[-1]
        return {}


class logger_mocked(object):
    def __init__(self):
        self.dbgmsg = []
        self.infomsg = []
        self.warnmsg = []
        self.errmsg = []

    def debug(self, *args):
        self.dbgmsg.extend(args)

    def info(self, *args):
        self.infomsg.extend(args)

    @deprecated(since='2020-09-23', message=(
        'The logging.warn method has been deprecated since Python 3.3.'
        'Use the warning method instead.'
    ))
    def warn(self, *args):
        self.warnmsg.extend(args)

    def warning(self, *args):
        self.warnmsg.extend(args)

    def error(self, *args):
        self.errmsg.extend(args)

    def __call__(self):
        return self


class CurrentActorMocked(object):  # pylint:disable=R0904
    def __init__(self, arch=architecture.ARCH_X86_64, envars=None, kernel='3.10.0-957.43.1.el7.x86_64',
                 release_id='rhel', src_ver='7.8', dst_ver='8.1', msgs=None, flavour='default'):
        envarsList = [EnvVar(name=k, value=v) for k, v in envars.items()] if envars else []
        version = namedtuple('Version', ['source', 'target'])(src_ver, dst_ver)
        release = namedtuple('OS_release', ['release_id', 'version_id'])(release_id, src_ver)

        self._common_folder = '../../files'
        self._common_tools_folder = '../../tools'
        self._actor_folder = 'files'
        self._actor_tools_folder = 'tools'
        self.configuration = namedtuple(
            'configuration', ['architecture', 'kernel', 'leapp_env_vars', 'os_release', 'version', 'flavour']
        )(arch, kernel, envarsList, release, version, flavour)
        self._msgs = msgs or []

    def __call__(self):
        return self

    def get_common_folder_path(self, folder):
        return os.path.join(self._common_folder, folder)

    def get_common_tool_path(self, name):
        return os.path.join(self._common_tools_folder, name)

    def consume(self, model):
        return iter(filter(  # pylint:disable=W0110,W1639
            lambda msg: isinstance(msg, model), self._msgs
        ))

    @property
    def log(self):
        return logging.getLogger(__name__)

    def get_actor_file_path(self, name):
        return os.path.join(self._actor_folder, name)

    # other functions and properties from the API - can be implemented as needed

    def serialize(self):
        raise NotImplementedError

    def get_answers(self, dialog):
        raise NotImplementedError

    def show_message(self, message):
        raise NotImplementedError

    @property
    def actor_files_paths(self):
        raise NotImplementedError

    @property
    def files_paths(self):
        raise NotImplementedError

    @property
    def common_files_paths(self):
        raise NotImplementedError

    @property
    def actor_tools_paths(self):
        raise NotImplementedError

    @property
    def common_tools_paths(self):
        raise NotImplementedError

    @property
    def tools_paths(self):
        raise NotImplementedError

    def get_folder_path(self, name):
        raise NotImplementedError

    def get_actor_folder_path(self, name):
        raise NotImplementedError

    def get_file_path(self, name):
        raise NotImplementedError

    def get_common_file_path(self, name):
        raise NotImplementedError

    def get_tool_path(self, name):
        raise NotImplementedError

    def get_actor_tool_path(self, name):
        raise NotImplementedError

    def run(self, *args):
        raise NotImplementedError

    def produce(self, *models):
        raise NotImplementedError

    def report_error(self, message, severity, details):
        raise NotImplementedError


def make_IOError(error):
    """
    Create an IOError instance

    Create an IOError instance with the given error number.

    :param error: the error number, e.g. errno.ENOENT
    """
    return IOError(error, os.strerror(error))


def make_OSError(error):
    """
    Create an OSError instance

    Create an OSError instance with the given error number.

    :param error: the error number, e.g. errno.ENOENT
    """
    return OSError(error, os.strerror(error))



#
# Functions for probing SCTP usage by DLM
#
import re

from leapp.libraries.common import utils
from leapp.libraries.stdlib import api


def check_dlm_cfgfile(_open=open):
    """Parse DLM config file.
    :param _open: object behind opening a file. Might be replaced
        by mocked one for the purpose of testing
    """
    fname = '/etc/dlm/dlm.conf'

    try:
        with _open(fname, 'r') as fp:
            cfgs = '[dlm]\n' + fp.read()
    except (OSError, IOError):
        return False

    cfg = utils.parse_config(cfgs)

    if not cfg.has_option('dlm', 'protocol'):
        return False

    proto = cfg.get('dlm', 'protocol').lower()
    return proto in ['sctp', 'detect', '1', '2']


def check_dlm_sysconfig(_open=open):
    """Parse /etc/sysconfig/dlm
    :param _open: object behind opening a file. Might be replaced
        by mocked one for the purpose of testing
    """

    regex = re.compile('^[^#]*DLM_CONTROLD_OPTS.*=.*(?:--protocol|-r)[ =]*([^"\' ]+).*', re.IGNORECASE)

    try:
        with _open('/etc/sysconfig/dlm', 'r') as fp:
            lines = fp.readlines()
    except (OSError, IOError):
        return False

    for line in lines:
        if regex.match(line):
            proto = regex.sub('\\1', line).lower().strip()
            if proto in ['sctp', 'detect']:
                return True

    return False


def is_dlm_using_sctp():
    if check_dlm_cfgfile():
        api.current_logger().info('DLM is configured to use SCTP on dlm.conf.')
        return True

    if check_dlm_sysconfig():
        api.current_logger().info('DLM is configured to use SCTP on sysconfig.')
        return True

    return False



from leapp.models import BlackListCA, BlackListError, Report
from leapp.utils.report import is_inhibitor


def test_actor_execution_empty(current_actor_context):
    current_actor_context.feed()
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_error_entry(current_actor_context):
    current_actor_context.feed(
        BlackListError(
            sourceDir="/blacklist",
            targetDir="/blocklist",
            error="Can't list /blacklist"
        )
    )
    current_actor_context.run()
    r = current_actor_context.consume(Report)
    assert r
    assert 'Could not access blacklist directory' in r[0].report['title']
    assert is_inhibitor(r[0].report)


def test_actor_single_entry(current_actor_context):
    current_actor_context.feed(
        BlackListCA(
            source="/blacklist/badcert.ca",
            sourceDir="/blacklist",
            target="/blocklist/badcert.ca",
            targetDir="/blocklist"
        )
    )
    current_actor_context.run()
    r = current_actor_context.consume(Report)
    assert r
    assert ('/blacklist/badcert.ca will be moved to /blocklist and '
            '/blacklist will be deleted') in r[0].report['summary']


def test_actor_two_entries_one_directory(current_actor_context):
    current_actor_context.feed(
        BlackListCA(
            source="/blacklist/badcert.ca",
            sourceDir="/blacklist",
            target="/blocklist/badcert.ca",
            targetDir="/blocklist"
        ),
        BlackListCA(
            source="/blacklist/badcert2.ca",
            sourceDir="/blacklist",
            target="/blocklist/badcert2.ca",
            targetDir="/blocklist"
        ),
    )
    current_actor_context.run()
    r = current_actor_context.consume(Report)
    assert r
    assert ('/blacklist/badcert.ca and /blacklist/badcert2.ca '
            'will be moved to /blocklist and /blacklist will '
            'be deleted') in r[0].report['summary']


def test_actor_three_entries_one_directory(current_actor_context):
    current_actor_context.feed(
        BlackListCA(
            source="/blacklist/badcert.ca",
            sourceDir="/blacklist",
            target="/blocklist/badcert.ca",
            targetDir="/blocklist"
        ),
        BlackListCA(
            source="/blacklist/badcert2.ca",
            sourceDir="/blacklist",
            target="/blocklist/badcert2.ca",
            targetDir="/blocklist"
        ),
        BlackListCA(
            source="/blacklist/badcert3.ca",
            sourceDir="/blacklist",
            target="/blocklist/badcert3.ca",
            targetDir="/blocklist"
        ),
    )
    current_actor_context.run()
    r = current_actor_context.consume(Report)
    assert r
    assert ('/blacklist/badcert.ca, /blacklist/badcert2.ca and '
            '/blacklist/badcert3.ca will be moved to /blocklist and '
            '/blacklist will be deleted') in r[0].report['summary']


def test_actor_two_entries_two_directories(current_actor_context):
    current_actor_context.feed(
        BlackListCA(
            source="/blacklist/badcert.ca",
            sourceDir="/blacklist",
            target="/blocklist/badcert.ca",
            targetDir="/blocklist"
        ),
        BlackListCA(
            source="/private/blacklist/badcert2.ca",
            sourceDir="/private/blacklist",
            target="/private/blocklist/badcert2.ca",
            targetDir="/private/blocklist"
        )
    )
    current_actor_context.run()
    r = current_actor_context.consume(Report)
    assert r
    assert ('/blacklist/badcert.ca will be moved to /blocklist '
            'and /private/blacklist/badcert2.ca will be moved to '
            '/private/blocklist and /blacklist and /private/blacklist '
            'will be deleted') in r[0].report['summary']


def test_actor_three_entries_tree_directories(current_actor_context):
    current_actor_context.feed(
        BlackListCA(
            source="/blacklist/badcert.ca",
            sourceDir="/blacklist",
            target="/blocklist/badcert.ca",
            targetDir="/blocklist"
        ),
        BlackListCA(
            source="/private/blacklist/badcert2.ca",
            sourceDir="/private/blacklist",
            target="/private/blocklist/badcert2.ca",
            targetDir="/private/blocklist"
        ),
        BlackListCA(
            source="/public/blacklist/badcert3.ca",
            sourceDir="/public/blacklist",
            target="/public/blocklist/badcert3.ca",
            targetDir="/public/blocklist"
        )
    )
    current_actor_context.run()
    r = current_actor_context.consume(Report)
    assert r
    assert ('/blacklist/badcert.ca will be moved to /blocklist, '
            '/private/blacklist/badcert2.ca will be moved to '
            '/private/blocklist and /public/blacklist/badcert3.ca '
            'will be moved to /public/blocklist and /blacklist, '
            '/private/blacklist and /public/blacklist will be deleted') in r[0].report['summary']



from leapp.actors import Actor
from leapp.libraries.actor import trustedgpgkeys
from leapp.models import InstalledRPM, TrustedGpgKeys
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class TrustedGpgKeysScanner(Actor):
    """
    Scan for trusted GPG keys.

    These include keys readily available in the source RPM DB, keys for N+1
    Red Hat release and custom keys stored in the trusted directory.
    """

    name = 'trusted_gpg_keys_scanner'
    consumes = (InstalledRPM,)
    produces = (TrustedGpgKeys,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        trustedgpgkeys.process()



from leapp.actors import Actor
from leapp.libraries.actor import gpgpubkeycheck
from leapp.models import TrustedGpgKeys
from leapp.reporting import Report
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class GpgPubkeyCheck(Actor):
    """
    Checks no unexpected GPG keys were installed during the upgrade.

    This should be mostly sanity check and this should not happen
    unless something went very wrong, regardless the gpgcheck was
    used (default) or not (with --no-gpgcheck option).
    """

    name = 'gpg_pubkey_check'
    consumes = (TrustedGpgKeys,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ApplicationsPhaseTag,)

    def process(self):
        gpgpubkeycheck.process()



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class OpenSshPermitRootLogin(Model):
    topic = SystemInfoTopic

    value = fields.StringEnum(['yes', 'prohibit-password',
                               'forced-commands-only', 'no'])
    """ Value of a PermitRootLogin directive. """
    in_match = fields.Nullable(fields.List(fields.String()))
    """ Criteria of Match blocks the PermitRootLogin directive occurred in, if any. """


class OpenSshConfig(Model):
    """
    OpenSSH server configuration.

    This model contains the first effective configuration option specified
    in the configuration file or a list of all the options specified
    in all the conditional blocks used throughout the file.
    """
    topic = SystemInfoTopic

    permit_root_login = fields.List(fields.Model(OpenSshPermitRootLogin))
    """ All PermitRootLogin directives. """
    use_privilege_separation = fields.Nullable(fields.StringEnum(['sandbox',
                                                                  'yes',
                                                                  'no']))
    """ Value of the UsePrivilegeSeparation directive, if present. Removed in RHEL 8. """
    protocol = fields.Nullable(fields.String())
    """ Value of the Protocols directive, if present. Removed in RHEL 8. """
    ciphers = fields.Nullable(fields.String())
    """ Value of the Ciphers directive, if present. Ciphers separated by comma. """
    macs = fields.Nullable(fields.String())
    """ Value of the MACs directive, if present. """
    deprecated_directives = fields.List(fields.String())
    """ Configuration directives that were deprecated in the new version of openssh. """
    subsystem_sftp = fields.Nullable(fields.String())
    """ The "Subsystem sftp" configuration option, if present """

    modified = fields.Boolean(default=False)
    """ True if the configuration file was modified. """



import errno
import os

from leapp.libraries.actor import vsftpdconfigread
from leapp.libraries.common.testutils import make_IOError, make_OSError
from leapp.models import DistributionSignedRPM, RPM


class MockFileOperations(object):
    def __init__(self):
        self.files = {}
        self.files_read = {}
        self.read_called = 0

    def _increment_read_counters(self, path):
        self.read_called += 1
        self.files_read.setdefault(path, 0)
        self.files_read[path] += 1

    def read(self, path):
        self._increment_read_counters(path)
        try:
            return self.files[path]
        except KeyError:
            raise make_IOError(errno.ENOENT)


class MockListDir(object):
    def __init__(self, path=None, file_names=None, to_raise=None):
        self.path = None if path is None else os.path.normpath(path)
        self.file_names = file_names
        self.to_raise = to_raise
        self.error = False

    def listdir(self, path):
        if self.to_raise:
            raise self.to_raise
        if os.path.normpath(path) == self.path:
            return self.file_names

        self.error = True
        raise make_OSError(errno.ENOENT)


def test_parse_config():
    content = 'anonymous_enable=YES'
    path = 'my_file'

    parsed = vsftpdconfigread._parse_config(path, content)

    assert parsed['anonymous_enable'] is True


def test_parsing_bad_config_gives_None():
    content = 'foo'
    path = 'my_file'

    parsed = vsftpdconfigread._parse_config(path, content)

    assert parsed is None


def test_get_parsed_configs():
    directory = '/etc/vsftpd'
    file_names = ['vsftpd.conf', 'foo.conf']
    listdir = MockListDir(directory, file_names)
    fileops = MockFileOperations()
    fileops.files[os.path.join(directory, file_names[0])] = 'anonymous_enable=YES\n' \
                                                            'ca_certs_file=/foo/bar\n'
    fileops.files[os.path.join(directory, file_names[1])] = 'anonymous_enable=NO\n'

    parsed_configs = list(vsftpdconfigread._get_parsed_configs(read_func=fileops.read,
                                                               listdir=listdir.listdir))

    assert not listdir.error
    assert len(fileops.files_read) == 2
    assert os.path.join(directory, file_names[0]) in fileops.files_read
    assert os.path.join(directory, file_names[1]) in fileops.files_read
    assert len(parsed_configs) == 2
    if parsed_configs[0][0] != os.path.join(directory, file_names[0]):
        parsed_configs.reverse()
    assert (os.path.join(directory, file_names[0]), {'anonymous_enable': True,
                                                     'ca_certs_file': '/foo/bar'}) in parsed_configs
    assert (os.path.join(directory, file_names[1]), {'anonymous_enable': False}) in parsed_configs


def test_get_parsed_configs_empty_dir():
    directory = '/etc/vsftpd'
    listdir = MockListDir(directory, [])
    fileops = MockFileOperations()

    parsed_configs = vsftpdconfigread._get_parsed_configs(read_func=fileops.read,
                                                          listdir=listdir.listdir)

    assert not listdir.error
    assert fileops.read_called == 0
    assert not parsed_configs


def test_get_parsed_configs_nonexistent_dir():
    listdir = MockListDir(to_raise=make_OSError(errno.ENOENT))
    fileops = MockFileOperations()

    parsed_configs = vsftpdconfigread._get_parsed_configs(read_func=fileops.read,
                                                          listdir=listdir.listdir)

    assert fileops.read_called == 0
    assert not parsed_configs


def test_get_parsed_configs_inaccessible_dir():
    listdir = MockListDir(to_raise=make_OSError(errno.EACCES))
    fileops = MockFileOperations()

    parsed_configs = vsftpdconfigread._get_parsed_configs(read_func=fileops.read,
                                                          listdir=listdir.listdir)

    assert fileops.read_called == 0
    assert not parsed_configs


def test_get_vsftpd_facts():
    directory = '/etc/vsftpd'
    file_names = ['vsftpd.conf', 'foo.conf', 'bar.conf']
    listdir = MockListDir(directory, file_names)
    fileops = MockFileOperations()
    fileops.files[os.path.join(directory, file_names[0])] = 'anonymous_enable=YES\n' \
                                                            'ca_certs_file=/foo/bar\n'
    fileops.files[os.path.join(directory, file_names[1])] = 'anonymous_enable=NO\n' \
                                                            'tcp_wrappers=YES\n'
    fileops.files[os.path.join(directory, file_names[2])] = 'strict_ssl_read_eof=yes\n' \
                                                            'tcp_wrappers=no\n'

    facts = vsftpdconfigread.get_vsftpd_facts(read_func=fileops.read, listdir=listdir.listdir)

    assert facts.default_config_hash == '892bae7b69eb66ec16afe842a15e53a5242155a4'
    assert len(facts.configs) == 3
    used_indices = set()
    for config in facts.configs:
        assert os.path.dirname(config.path) == directory
        file_name = os.path.basename(config.path)
        ix = file_names.index(file_name)
        if ix in used_indices:
            assert False
        used_indices.add(ix)
        if ix == 0:
            assert config.strict_ssl_read_eof is None
            assert config.tcp_wrappers is None
        elif ix == 1:
            assert config.strict_ssl_read_eof is None
            assert config.tcp_wrappers is True
        elif ix == 2:
            assert config.strict_ssl_read_eof is True
            assert config.tcp_wrappers is False
        else:
            assert False


def test_get_vsftpd_facts_empty_dir():
    listdir = MockListDir('/etc/vsftpd', [])
    fileops = MockFileOperations()

    facts = vsftpdconfigread.get_vsftpd_facts(read_func=fileops.read, listdir=listdir.listdir)

    assert facts.default_config_hash is None
    assert not facts.configs


def test_get_vsftpd_facts_nonexistent_dir():
    listdir = MockListDir(to_raise=make_OSError(errno.ENOENT))
    fileops = MockFileOperations()

    facts = vsftpdconfigread.get_vsftpd_facts(read_func=fileops.read, listdir=listdir.listdir)

    assert facts.default_config_hash is None
    assert not facts.configs


def test_get_vsftpd_facts_inaccessible_dir():
    listdir = MockListDir(to_raise=make_OSError(errno.EACCES))
    fileops = MockFileOperations()

    facts = vsftpdconfigread.get_vsftpd_facts(read_func=fileops.read, listdir=listdir.listdir)

    assert facts.default_config_hash is None
    assert not facts.configs


def test_is_processable_vsftpd_installed():
    installed_rpms = [
        RPM(name='sendmail', version='8.14.7', release='5.el7', epoch='0',
            packager='foo', arch='x86_64', pgpsig='bar'),
        RPM(name='vsftpd', version='3.0.2', release='25.el7', epoch='0',
            packager='foo', arch='x86_64', pgpsig='bar'),
        RPM(name='postfix', version='2.10.1', release='7.el7', epoch='0',
            packager='foo', arch='x86_64', pgpsig='bar')]
    installed_rpm_facts = DistributionSignedRPM(items=installed_rpms)

    res = vsftpdconfigread.is_processable(installed_rpm_facts)

    assert res is True


def test_is_processable_vsftpd_not_installed():
    installed_rpms = [
        RPM(name='sendmail', version='8.14.7', release='5.el7', epoch='0',
            packager='foo', arch='x86_64', pgpsig='bar'),
        RPM(name='postfix', version='2.10.1', release='7.el7', epoch='0',
            packager='foo', arch='x86_64', pgpsig='bar')]
    installed_rpm_facts = DistributionSignedRPM(items=installed_rpms)

    res = vsftpdconfigread.is_processable(installed_rpm_facts)

    assert res is False



import json
import os
from collections import defaultdict, namedtuple
from enum import IntEnum
from itertools import chain

from leapp import reporting
from leapp.exceptions import StopActorExecution
from leapp.libraries.common import fetch
from leapp.libraries.common.config import architecture
from leapp.libraries.common.rpms import get_leapp_packages, LeappComponents
from leapp.libraries.stdlib import api

# NOTE(mhecko): The modulestream field contains a set of modulestreams until the very end when we generate a Package
# for every modulestream in this set.
_Package = namedtuple('Package', ['name',         # str
                                  'repository',   # str
                                  'modulestream'  # (str, str) or None
                                  ])


class Package(_Package):  # noqa: E0102; pylint: disable=function-redefined
    def __repr__(self):
        ms = ''
        if self.modulestream:
            # The package class contains multiple modulestreams until a package for every single of these modulestreams
            # is generated. We have to check whether modulestream in fact contains multiple modulestreams and handle it
            # accordingly, so that we print Package when debugging.
            if len(self.modulestream) == 2 and all((isinstance(item, str) for item in self.modulestream)):
                ms = '@{0}:{1}'.format(*self.modulestream)
            else:
                ms = '{{{0}}}'.format(','.join(str(item) for item in self.modulestream))
        return '{n}:{r}{ms}'.format(n=self.name, r=self.repository, ms=ms)

    def __hash__(self):
        return hash((self.name, self.modulestream))

    def __eq__(self, other):
        return (self.name, self.modulestream) == (other.name, other.modulestream)


Event = namedtuple('Event', ['id',            # int
                             'action',        # An instance of Action
                             'in_pkgs',       # A set of Package named tuples
                             'out_pkgs',      # A set of Package named tuples
                             'from_release',  # A tuple representing a release in format (major, minor)
                             'to_release',    # A tuple representing a release in format (major, minor)
                             'architectures'  # A list of strings representing architectures
                             ])


class Action(IntEnum):
    PRESENT = 0
    REMOVED = 1
    DEPRECATED = 2
    REPLACED = 3
    SPLIT = 4
    MERGED = 5
    MOVED = 6
    RENAMED = 7


def get_pes_events(pes_json_directory, pes_json_filename):
    """
    Get all the events from the source JSON file exported from PES.

    :return: List of Event tuples, where each event contains event type and input/output pkgs
    """
    try:
        # NOTE(pstodulk): load_data_assert raises StopActorExecutionError, see
        # the code for more info. Keeping the handling on the framework in such
        # a case as we have no work to do in such a case here.
        events_data = fetch.load_data_asset(api.current_actor(),
                                            pes_json_filename,
                                            asset_fulltext_name='PES events file',
                                            docs_url='',
                                            docs_title='')
        if not events_data:
            return None

        if not events_data.get('packageinfo'):
            raise ValueError('Found PES data with invalid structure')

        all_events = list(chain(*[parse_entry(entry) for entry in events_data['packageinfo']]))
        arch = api.current_actor().configuration.architecture
        events_matching_arch = [e for e in all_events if not e.architectures or arch in e.architectures]
        return events_matching_arch
    except (ValueError, KeyError):
        local_path = os.path.join(pes_json_directory, pes_json_filename)
        title = 'Missing/Invalid PES data file ({})'.format(local_path)
        summary = (
            'All official data files are nowadays part of the installed rpms.'
            ' This issue is usually encountered when the data files are incorrectly customized, replaced, or removed'
            ' (e.g. by custom scripts).'
        )
        hint = (
            ' In case you want to recover the original {lp} file, remove it (if it still exists)'
            ' and reinstall the following rpms: {rpms}.'
            .format(
                lp=local_path,
                rpms=', '.join(get_leapp_packages(component=LeappComponents.REPOSITORY))
            )
        )
        reporting.create_report([
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Remediation(hint=hint),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.SANITY, reporting.Groups.INHIBITOR]),
            reporting.RelatedResource('file', os.path.join(pes_json_directory, pes_json_filename))
        ])
        raise StopActorExecution()


def generate_event_for_ms_mapping_entry(from_ms_to_ms_entry, event):
    from_modulestream, to_modulestreams = from_ms_to_ms_entry

    in_pkgs_matching_from_ms = {
        Package(p.name, p.repository, from_modulestream) for p in event.in_pkgs if from_modulestream in p.modulestream
    }

    # Out pkgs are a bit harder to pick, as there are more than one modulestreams in to_modulestreams
    out_pkgs_matching_to_ms = set()
    for pkg in event.out_pkgs:
        target_modulestreams = set(pkg.modulestream).intersection(to_modulestreams)
        if len(target_modulestreams) > 1:
            # This is not expected to happen. Produce a warning, but continue by picking a single of the modulestreams.
            api.current_logger().warning('Event {0} - output package {1} matches multiple to_modulestreams '
                                         ' when generating events per modulestream mapping.')
        if not target_modulestreams:
            # No target_modulestreams mean that this package should not be included in the resulting event
            continue

        target_modulestream = next(iter(target_modulestreams))
        out_pkgs_matching_to_ms.add(Package(pkg.name, pkg.repository, target_modulestream))

    return Event(
        event.id,
        event.action,
        in_pkgs_matching_from_ms,
        out_pkgs_matching_to_ms,
        event.from_release,
        event.to_release,
        event.architectures
    )


def event_by_modulestream_mapping(mapping, event):
    """
    Generate events with matching event.ID according to the the given modulestream mapping.

    Modulestream mapping allows for a compact symbolic representation of multiple events having the same input/output
    packages differing only in the modules they are in. For example consider event renaming package X with module
    streams [MSX0, MSX1], to a new package with the name Y and possible module streams [MSY0, MSY1]. The attached
    modulestream map {MSX0 -> MSY0, MSX1 -> MSY1} would mean that the renaming event represents two events --- one
    renaming (X, ms=MSX0) to (Y, ms=MSY0), and the second one renaming (X, ms=MSX1) to (Y, ms=MSY1).

    :param mapping: A dictionary mapping package's input modulestream to its output modulestreams.
    :param event: PES event that should be uncompressed.
    :returns: A list of generated PES events.
    """
    if not mapping:
        # In case there is no mapping, assume all is going to non modular
        mapping = defaultdict(set)
        for package in event.in_pkgs:
            for ms in package.modulestream or ():
                mapping[ms].add(None)

    return [generate_event_for_ms_mapping_entry(from_ms_to_ms_entry, event) for from_ms_to_ms_entry in mapping.items()]


def parse_pes_events(json_data):
    """
    Parse JSON data returning PES events

    :return: List of Event tuples, where each event contains event type and input/output pkgs
    """
    data = json.loads(json_data)
    if not isinstance(data, dict) or not data.get('packageinfo'):
        raise ValueError('Found PES data with invalid structure')

    return list(chain(*[parse_entry(entry) for entry in data['packageinfo']]))


def parse_entry(entry):
    """
    Parse PES event data

    :param entry: A dict with the following structure:
                  {
                      'action': 3,
                      'id': 15,
                      'initial_release': {  # can be None
                          'z_stream': None,
                          'major_version': 7,
                          'tag': None,
                          'os_name': 'RHEL',
                          'minor_version': 7
                      },
                      'release': {
                          'z_stream': None,
                          'major_version': 8,
                          'tag': None,
                          'os_name': 'RHEL',
                          'minor_version': 0
                      },
                      'in_packageset': {
                          'set_id': 20,
                          'package': [
                              {
                                  'name': 'espeak',
                                  'repository': 'rhel7-optional'
                              }
                          ]
                      },
                      'out_packageset': {  # can be None
                          'set_id': 21,
                          'package': [
                              {
                                  'name': 'espeak-ng',
                                  'repository': 'rhel8-AppStream'
                              }
                          ]
                      },
                      'architectures': [  # can be empty
                          'x86_64',
                          'aarch64',
                          'ppc64le',
                          's390x'
                      ]
                  }
    """

    event_id = entry.get('id') or 0

    action_id = entry['action']
    if action_id < 0 or action_id >= len(Action):
        raise ValueError('Found event with invalid action ID: {}'.format(action_id))
    action = Action(action_id)

    in_pkgs = parse_packageset(entry.get('in_packageset') or {})
    out_pkgs = parse_packageset(entry.get('out_packageset') or {})

    # parse_release handles missing release data, no need to supply a default value
    from_release = parse_release(entry.get('initial_release'))
    to_release = parse_release(entry.get('release'))

    architectures = entry.get('architectures') or []
    invalid_archs = tuple(arch for arch in architectures if arch not in architecture.ARCH_ACCEPTED)
    if invalid_archs:
        raise ValueError('Found event with invalid architecture{0}: {1}'.format('s' if len(invalid_archs) > 1 else '',
                                                                                ', '.join(invalid_archs)))

    # Parse modulestream maps
    modulestream_map_entry = entry.get('modulestream_maps') or []

    modulestream_maps = defaultdict(set)
    for mapping_entry in modulestream_map_entry:
        in_ms_entry = mapping_entry.get('in_modulestream', {})
        in_modulestream = (in_ms_entry.get('name'), in_ms_entry.get('stream')) if in_ms_entry else None

        out_ms_entry = mapping_entry.get('out_modulestream', {})
        out_modulestream = (out_ms_entry.get('name'), out_ms_entry.get('stream')) if out_ms_entry else None

        # One modulestream might have more than one output modulestream - eg. the output packages of split belong to
        # different modulestreams: (pkg_in, ms1) is split to (pkg_out1, ms2), (pkg_out2, ms3). Therefore, we need to
        # collect for every input modulestream all the output modulestreams it maps to.
        modulestream_maps[in_modulestream].add(out_modulestream)

    return event_by_modulestream_mapping(
        modulestream_maps,
        Event(event_id, action, in_pkgs, out_pkgs, from_release, to_release, architectures)
    )


def parse_packageset(packageset):
    """
    Get "input" or "output" packages and their repositories from each PES event.

    :return: set of Package tuples
    """
    packageset_pkgs = packageset.get('package', packageset.get('packages', []))
    parsed_pkgs = set()

    for package in packageset_pkgs:
        # The package can have either the modulestreams field (a list of modulestream entries)
        # or a modulestream entry containing a single module stream
        modulestream_entries = package.get('modulestreams')
        if not modulestream_entries:
            modulestream_entry = package.get('modulestream')
            if isinstance(modulestream_entry, dict):
                # There is only one module stream present
                modulestream_entries = (modulestream_entry,)

        if not modulestream_entries:
            modulestreams = (None, )
        else:
            modulestreams = tuple((ms['name'], ms['stream']) if ms else None for ms in modulestream_entries)

        pkg = Package(package['name'], package['repository'], modulestreams)
        parsed_pkgs.add(pkg)

    return parsed_pkgs


def parse_release(release):
    return (release['major_version'], release['minor_version']) if release else (0, 0)



import os

from leapp.models import FirewalldGlobalConfig


def read_config():
    default_conf = FirewalldGlobalConfig()

    path = '/etc/firewalld/firewalld.conf'
    if not os.path.exists(path):
        return default_conf

    conf_dict = {}
    with open(path) as conf_file:
        for line in conf_file:
            (key, _unused, value) = line.partition('=')
            if not value:
                continue

            value = value.lower().strip()
            if value in ['yes', 'true']:
                value = True
            if value in ['no', 'false']:
                value = False

            # Only worry about config used by our Model
            key = key.lower().strip()
            if hasattr(default_conf, key):
                conf_dict[key] = value

    return FirewalldGlobalConfig(**conf_dict)



from leapp.actors import Actor
from leapp.libraries.actor import ifcfgscanner
from leapp.models import IfCfg
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class IfCfgScanner(Actor):
    """
    Scan ifcfg files with legacy network configuration
    """

    name = "ifcfg_scanner"
    consumes = ()
    produces = (IfCfg,)
    tags = (IPUWorkflowTag, FactsPhaseTag,)

    def process(self):
        ifcfgscanner.process()



from leapp import reporting
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM

# Summary for postgresql-server report
report_server_inst_summary = (
    'PostgreSQL server component will be upgraded. Since RHEL-8 includes'
    ' PostgreSQL server 10 by default, which is incompatible with 9.2'
    ' included in RHEL-7, it is necessary to proceed with additional steps'
    ' for the complete upgrade of the PostgreSQL data.'
)

report_server_inst_hint = (
    'Back up your data before proceeding with the upgrade'
    ' and follow steps in the documentation section "Migrating to a RHEL 8 version of PostgreSQL"'
    ' after the upgrade.'
)

# Link URL for postgresql-server report
report_server_inst_link_url = 'https://red.ht/rhel-8-migrate-postgresql-server'

# List of dropped extensions from postgresql-contrib package
report_contrib_inst_dropext = ['dummy_seclabel', 'test_parser', 'tsearch2']

# Summary for postgresql-contrib report
report_contrib_inst_summary = (
    'Please note that some extensions have been dropped from the'
    ' postgresql-contrib package and might not be available after'
    ' the upgrade:{}'
    .format(''.join(['\n    - {}'.format(i) for i in report_contrib_inst_dropext]))
)


def _report_server_installed():
    """
    Create report on postgresql-server package installation detection.

    Should remind user about present PostgreSQL server package
    installation, warn them about necessary additional steps, and
    redirect them to online documentation for the upgrade process.
    """
    reporting.create_report([
        reporting.Title('PostgreSQL (postgresql-server) has been detected on your system'),
        reporting.Summary(report_server_inst_summary),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.ExternalLink(title='Migrating to a RHEL 8 version of PostgreSQL',
                               url=report_server_inst_link_url),
        reporting.RelatedResource('package', 'postgresql-server'),
        reporting.Remediation(hint=report_server_inst_hint),
        ])


def _report_contrib_installed():
    """
    Create report on postgresql-contrib package installation detection.

    Should remind user about present PostgreSQL contrib package
    installation and provide them with a list of extensions no longer
    shipped with this package.
    """
    reporting.create_report([
        reporting.Title('PostgreSQL (postgresql-contrib) has been detected on your system'),
        reporting.Summary(report_contrib_inst_summary),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.RelatedResource('package', 'postgresql-contrib')
        ])


def report_installed_packages(_context=api):
    """
    Create reports according to detected PostgreSQL packages.

    Create the report if the postgresql-server rpm (RH signed) is installed.
    Additionally, create another report if the postgresql-contrib rpm
    is installed.
    """
    has_server = has_package(DistributionSignedRPM, 'postgresql-server', context=_context)
    has_contrib = has_package(DistributionSignedRPM, 'postgresql-contrib', context=_context)

    if has_server:
        # postgresql-server
        _report_server_installed()
        if has_contrib:
            # postgresql-contrib
            _report_contrib_installed()



from leapp.actors import Actor
from leapp.libraries.actor import removeobsoleterpmgpgkeys
from leapp.models import DNFWorkaround, InstalledRPM
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class RemoveObsoleteGpgKeys(Actor):
    """
    Remove obsoleted RPM GPG keys.

    New version might make existing RPM GPG keys obsolete. This might be caused
    for example by the hashing algorithm becoming deprecated or by the key
    getting replaced.

    A DNFWorkaround is registered to actually remove the keys.
    """

    name = "remove_obsolete_gpg_keys"
    consumes = (InstalledRPM,)
    produces = (DNFWorkaround,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        removeobsoleterpmgpgkeys.process()



from leapp import reporting
from leapp.libraries.stdlib import api

VDO_DOC_URL = 'https://red.ht/import-existing-vdo-volumes-to-lvm'


def _report_skip_check():
    if not api.current_actor().get_vdo_answer():
        return

    summary = ('User has asserted all VDO devices on the system have been '
               'successfully converted to LVM management or no VDO '
               'devices are present.')
    reporting.create_report([
        reporting.Title('Skipping the VDO check of block devices'),
        reporting.Summary(summary),
        reporting.Severity(reporting.Severity.INFO),
        reporting.Groups([reporting.Groups.SERVICES, reporting.Groups.DRIVERS]),
    ])


def _process_failed_check_devices(conversion_info):
    # Post-conversion VDOs that were not successfully checked for having
    # completed the migration to LVM management.
    # Return True if failed checks detected
    devices = [x for x in conversion_info.post_conversion if (not x.complete) and x.check_failed]
    devices += [x for x in conversion_info.undetermined_conversion if x.check_failed]
    if not devices:
        return False

    if api.current_actor().get_vdo_answer():
        # User asserted all possible VDO should be already converted - skip
        return True

    names = [x.name for x in devices]
    multiple = len(names) > 1
    summary = ['Unexpected result checking device{0}'.format('s' if multiple else '')]
    summary.extend([x.failure for x in devices])
    summary = '\n'.join(summary)

    remedy_hint = ''.join(('Resolve the conditions leading to the reported '
                           'failure{0} '.format('s' if multiple else ''),
                           'and re-run the upgrade.'))

    reporting.create_report([
        reporting.Title('Checking VDO conversion to LVM management of block devices failed'),
        reporting.Summary(summary),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.SERVICES, reporting.Groups.DRIVERS]),
        reporting.Remediation(hint=remedy_hint),
        reporting.Groups([reporting.Groups.INHIBITOR]),
        reporting.ExternalLink(url=VDO_DOC_URL, title='Importing existing VDO volumes to LVM')
    ])
    return True


def _process_post_conversion_vdos(vdos):
    # Post-conversion VDOs that have definitively been shown to not have
    # completed the migration to LVM management generate an inhibiting report.
    post_conversion = [x for x in vdos if (not x.complete) and (not x.check_failed)]
    if post_conversion:
        devices = [x.name for x in post_conversion]
        multiple = len(devices) > 1
        summary = (
            'VDO device{s_suffix} \'{devices_str}\' '
            'did not complete migration to LVM management. '
            'The named device{s_suffix} {was_were} successfully converted '
            'at the device format level; however, the expected LVM management '
            'portion of the conversion did not take place. This indicates '
            'that an exceptional condition (for example, a system crash) '
            'likely occurred during the conversion process. The LVM portion '
            'of the conversion must be performed in order for upgrade '
            'to proceed.'
            .format(
                s_suffix='s' if multiple else '',
                devices_str=', '.join(devices),
                was_were='were' if multiple else 'was',
            )
        )

        remedy_hint = ('Consult the VDO to LVM conversion process '
                       'documentation for how to complete the conversion.')

        reporting.create_report([
            reporting.Title('Detected VDO devices that have not finished the conversion to LVM management.'),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.SERVICES, reporting.Groups.DRIVERS]),
            reporting.Remediation(hint=remedy_hint),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.ExternalLink(url=VDO_DOC_URL, title='Importing existing VDO volumes to LVM')
        ])


def _process_pre_conversion_vdos(vdos):
    # Pre-conversion VDOs generate an inhibiting report.
    if vdos:
        devices = [x.name for x in vdos]
        multiple = len(devices) > 1
        summary = (
            'VDO device{s_suffix} \'{devices_str}\' require{s_suffix_verb} '
            'migration to LVM management.'
            'After performing the upgrade VDO devices can only be '
            'managed via LVM. Any VDO device not currently managed '
            'by LVM must be converted to LVM management before '
            'upgrading. The data on any VDO device not converted to '
            'LVM management will be inaccessible after upgrading.'
            .format(
                s_suffix='s' if multiple else '',
                s_suffix_verb='' if multiple else 's',
                devices_str=', '.join(devices),
            )
        )

        remedy_hint = ('Consult the VDO to LVM conversion process '
                       'documentation for how to perform the conversion.')

        reporting.create_report([
            reporting.Title('Detected VDO devices not managed by LVM'),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.SERVICES, reporting.Groups.DRIVERS]),
            reporting.Remediation(hint=remedy_hint),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.ExternalLink(url=VDO_DOC_URL, title='Importing existing VDO volumes to LVM')
        ])


def _process_undetermined_conversion_devices(devices):
    # Practically we cannot have both unchecked and checked (with failure)
    # devices.  Devices are only unchecked if the vdo package is not installed
    # on the system and can only be checked with failure if it is.  Only in the
    # case of a bug could there be both.
    #
    # We process the two possibilities serially knowing that only one, if any,
    # could occur, but that if somehow both do (a bug) we're at least made
    # aware that it happened.
    #
    # A device can only end up as undetermined either via a check that failed
    # or if it was not checked.  If the info for the device indicates that it
    # did not have a check failure that means it was not checked.
    # Return True if failed checks detected

    unchecked = [x for x in devices if not x.check_failed]
    if not unchecked:
        return False

    if api.current_actor().get_vdo_answer():
        # User asserted no VDO devices are present
        return True

    summary = (
        'The check of block devices could not be performed as the \'vdo\' '
        'package is not installed. All VDO devices must be converted to '
        'LVM management prior to the upgrade to prevent the loss of data.')
    remedy_hint = ('Install the \'vdo\' package and re-run upgrade to '
                   'check for VDO devices requiring conversion or confirm '
                   'that all VDO devices, if any, are managed by LVM.')

    reporting.create_report([
        reporting.Title('Cannot perform the VDO check of block devices'),
        reporting.Summary(summary),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.SERVICES, reporting.Groups.DRIVERS]),
        reporting.Remediation(hint=remedy_hint),
        reporting.Groups([reporting.Groups.INHIBITOR]),
        reporting.ExternalLink(url=VDO_DOC_URL, title='Importing existing VDO volumes to LVM')
    ])
    return True


def check_vdo(conversion_info):
    _process_pre_conversion_vdos(conversion_info.pre_conversion)
    _process_post_conversion_vdos(conversion_info.post_conversion)

    detected_under_dev = _process_undetermined_conversion_devices(conversion_info.undetermined_conversion)
    detected_failed_check = _process_failed_check_devices(conversion_info)
    if detected_under_dev or detected_failed_check:
        _report_skip_check()



import os

from leapp.libraries.actor import multipathconfupdate
from leapp.libraries.common import multipathutil
from leapp.models import MultipathConfig, MultipathConfigOption

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


def build_config(val):
    all_devs_options_val = []
    for name_val, value_val in val[16]:
        option = MultipathConfigOption(name=name_val, value=value_val)
        all_devs_options_val.append(option)
    return MultipathConfig(
            pathname=val[0],
            default_path_checker=val[1],
            config_dir=val[2],
            default_retain_hwhandler=val[3],
            default_detect_prio=val[4],
            default_detect_checker=val[5],
            reassign_maps=val[6],
            hw_str_match_exists=val[7],
            ignore_new_boot_devs_exists=val[8],
            new_bindings_in_boot_exists=val[9],
            unpriv_sgio_exists=val[10],
            detect_path_checker_exists=val[11],
            overrides_hwhandler_exists=val[12],
            overrides_pg_timeout_exists=val[13],
            queue_if_no_path_exists=val[14],
            all_devs_section_exists=val[15],
            all_devs_options=all_devs_options_val)


default_rhel7_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/default_rhel7.conf'), 'directio', os.path.join(CUR_DIR, 'files/conf.d'),
     False, False, False, True, True, True, True, True, True, False, False, True, False, [], ])

all_devs_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/all_devs.conf'), None, None, None, None, None, None, False, False, False,
     True, True, False, False, True, True,
     [('path_checker', 'rdac'), ('detect_checker', 'yes'), ('features', '2 pg_init_retries 50'),
      ('path_selector', 'service-time 0'), ('fast_io_fail_tmo', '5'), ('no_path_retry', 'queue'), ], ])

empty_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/empty.conf'), None, None, None, None, None, None, False, False, False, False,
     False, False, False, False, False, [], ])

default_rhel8_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/default_rhel8.conf'), 'tur', '/etc/multipath/conf.d', True, True, None, False,
     False, False, False, False, False, False, False, False, False, [], ])

all_the_things_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/all_the_things.conf'), 'directio', os.path.join(CUR_DIR, 'files/conf.d'),
     False, False, False, True, True, True, True, True, True, True, True, True, True,
     [('no_path_retry', 'fail'), ('features', '0')], ])

already_updated_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/already_updated.conf'), None, os.path.join(CUR_DIR, 'files/conf.d'), None,
     None, None, None, False, False, False, False, False, False, False, False, False, [], ])

ugly1_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/ugly1.conf'), 'directio', os.path.join(CUR_DIR, 'files/conf.d'), False, False,
     False, True, True, True, True, True, True, True, True, True, True,
     [('dev_loss_tmo', '60'), ('path_selector', 'service-time 0')], ])

# same results as all_devs_conf
ugly2_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/ugly2.conf'), None, None, None, None, None, None, False, False, False, True,
     True, False, False, True, True,
     [('path_checker', 'rdac'), ('detect_checker', 'yes'), ('features', '2 pg_init_retries 50'),
      ('path_selector', 'service-time 0'), ('fast_io_fail_tmo', '5'), ('no_path_retry', 'queue'), ], ])

just_checker_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/just_checker.conf'), 'rdac', '/etc/multipath/conf.d', True, True, None, False,
     False, False, False, False, False, False, False, False, False, [], ])

just_detect_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/just_detect.conf'), None, None, None, False, None, None, False, False, False,
     False, False, False, False, False, False, [], ])

just_reassign_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/just_reassign.conf'), None, None, None, None, None, True, False, False, False,
     False, False, False, False, False, False, [], ])

just_exists_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/just_exists.conf'), None, None, None, None, None, None, False, False, False,
     False, False, False, False, True, False, [], ])

just_all_devs_conf = build_config(
    [os.path.join(CUR_DIR, 'files/before/just_all_devs.conf'), None, None, None, None, None, None, False, False, False,
     False, False, False, False, False, True, [], ])


def test_configs():
    tests = [(default_rhel7_conf, os.path.join(CUR_DIR, 'files/after/default_rhel7.conf'),),
             (all_devs_conf, os.path.join(CUR_DIR, 'files/after/all_devs.conf')), (empty_conf, None),
             (default_rhel8_conf, None),
             (all_the_things_conf, os.path.join(CUR_DIR, 'files/after/all_the_things.conf'),),
             (already_updated_conf, None), (ugly1_conf, os.path.join(CUR_DIR, 'files/after/ugly1.conf')),
             (ugly2_conf, os.path.join(CUR_DIR, 'files/after/ugly2.conf')),
             (just_checker_conf, os.path.join(CUR_DIR, 'files/after/just_checker.conf'),),
             (just_detect_conf, os.path.join(CUR_DIR, 'files/after/just_detect.conf'),),
             (just_reassign_conf, os.path.join(CUR_DIR, 'files/after/just_reassign.conf'),),
             (just_exists_conf, os.path.join(CUR_DIR, 'files/after/just_exists.conf'),),
             (just_all_devs_conf, os.path.join(CUR_DIR, 'files/after/just_all_devs.conf'),), ]
    for config, expected_config in tests:
        config_lines = multipathconfupdate._update_config(config)
        if config_lines is None:
            assert expected_config is None
            continue
        expected_lines = multipathutil.read_config(expected_config)
        assert expected_lines is not None
        assert len(config_lines) == len(expected_lines)
        for config_line, expected_line in zip(config_lines, expected_lines):
            assert config_line == expected_line



from leapp.actors import Actor
from leapp.libraries.actor.checkmountoptions import check_mount_options
from leapp.models import StorageInfo
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckMountOptions(Actor):
    """
    Check for mount options preventing the upgrade.

    Checks performed:
        - /var is mounted with the noexec option
    """
    name = "check_mount_options"
    consumes = (StorageInfo,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag,)

    def process(self):
        check_mount_options()



import functools
import os
import subprocess

import pyudev

from leapp import reporting
from leapp.libraries.stdlib import api
from leapp.models import (
    FstabEntry,
    LsblkEntry,
    LvdisplayEntry,
    MountEntry,
    PartitionEntry,
    PvsEntry,
    StorageInfo,
    SystemdMountEntry,
    VgsEntry
)


def aslist(f):
    """ Decorator used to convert generator to list """
    @functools.wraps(f)
    def inner(*args, **kwargs):
        return list(f(*args, **kwargs))
    return inner


def _is_file_readable(path):
    """ Verify if file exists and is readable """
    return os.path.isfile(path) and os.access(path, os.R_OK)


def _get_cmd_output(cmd, delim, expected_len):
    """ Verify if command exists and return output """
    if not any(os.access(os.path.join(path, cmd[0]), os.X_OK) for path in os.environ['PATH'].split(os.pathsep)):
        api.current_logger().warning("'%s': command not found" % cmd[0])
        return

    try:
        # FIXME: Will keep call to subprocess until our stdlib supports "env" parameter
        # when there is any fd except 0,1,2 open, lvm closes the fd and prints a warning.
        # In our case /dev/urandom has other fd opened, probably for caching purposes.
        output = subprocess.check_output(cmd, env={'LVM_SUPPRESS_FD_WARNINGS': '1', 'PATH': os.environ['PATH']})

    except subprocess.CalledProcessError as e:
        api.current_logger().debug("Command '%s' return non-zero exit status: %s" % (" ".join(cmd), e.returncode))
        return

    if bytes is not str:
        output = output.decode('utf-8')

    for entry in output.split('\n'):
        entry = entry.strip()
        if not entry:
            continue

        data = entry.split(delim)
        data.extend([''] * (expected_len - len(data)))

        yield data


@aslist
def _get_partitions_info(partitions_path):
    """ Collect storage info from /proc/partitions file """
    if _is_file_readable(partitions_path):
        with open(partitions_path, 'r') as partitions:
            skipped_header = False
            for entry in partitions:
                if entry.startswith('#'):
                    continue

                if not skipped_header:
                    skipped_header = True
                    continue

                entry = entry.strip()
                if not entry:
                    continue

                major, minor, blocks, name = entry.split()
                yield PartitionEntry(
                    major=major,
                    minor=minor,
                    blocks=blocks,
                    name=name)


@aslist
def _get_fstab_info(fstab_path):
    """ Collect storage info from /etc/fstab file """
    if _is_file_readable(fstab_path):
        with open(fstab_path, 'r') as fstab:
            for line, entry in enumerate(fstab, 1):
                if entry.startswith('#'):
                    continue

                entry = entry.strip()
                if not entry:
                    continue

                entries = entry.split()

                if len(entries) == 4:
                    entries.append('0')

                if len(entries) == 5:
                    entries.append('0')

                if len(entries) != 6:
                    if any(value.startswith('#') for value in entries):
                        remediation = (
                            'Comments in the /etc/fstab file must be at the beginning of the line, your file has a'
                            ' comment at the end of the line at line {}, please edit and fix this, for further'
                            ' information read fstab man page (man 5 fstab).'.format(line)
                        )
                    else:
                        remediation = (
                            'The /etc/fstab file must have at least 4 values and at most 6 per line, your file on the'
                            ' line: {} have {} values, please edit and fix this, for further information read'
                            ' fstab man page (man 5 fstab). '.format(line, len(entries))
                        )
                    summary = (
                        'The fstab configuration file seems to be invalid. You need to fix it to be able to proceed'
                        ' with the upgrade process.'
                    )
                    reporting.create_report([
                        reporting.Title('Problems with parsing data in /etc/fstab'),
                        reporting.Summary(summary),
                        reporting.Severity(reporting.Severity.HIGH),
                        reporting.Groups([reporting.Groups.FILESYSTEM]),
                        reporting.Groups([reporting.Groups.INHIBITOR]),
                        reporting.Remediation(hint=remediation),
                        reporting.RelatedResource('file', '/etc/fstab')
                    ])

                    api.current_logger().error(summary)
                    break

                # NOTE: fstab entries are yielded in the same order as in the /etc/fstab
                fs_spec, fs_file, fs_vfstype, fs_mntops, fs_freq, fs_passno = entries
                yield FstabEntry(
                    fs_spec=fs_spec,
                    fs_file=fs_file,
                    fs_vfstype=fs_vfstype,
                    fs_mntops=fs_mntops,
                    fs_freq=fs_freq,
                    fs_passno=fs_passno)


@aslist
def _get_mount_info(path):
    """ Collect storage info """
    with open(path, 'r') as fp:
        for line in [l.strip() for l in fp.readlines()]:
            device, mount, tp, options, _, _ = line.split(' ')
            yield MountEntry(
                name=device,
                mount=mount,
                tp=tp,
                options=options
            )


@aslist
def _get_lsblk_info():
    """ Collect storage info from lsblk command """
    cmd = ['lsblk', '-pbnr', '--output', 'NAME,MAJ:MIN,RM,SIZE,RO,TYPE,MOUNTPOINT']
    for entry in _get_cmd_output(cmd, ' ', 7):
        dev_path, maj_min, rm, bsize, ro, tp, mountpoint = entry
        lsblk_cmd = ['lsblk', '-nr', '--output', 'NAME,KNAME,SIZE', dev_path]
        lsblk_info_for_devpath = next(_get_cmd_output(lsblk_cmd, ' ', 3), None)
        if not lsblk_info_for_devpath:
            return

        name, kname, size = lsblk_info_for_devpath
        yield LsblkEntry(
            name=name,
            kname=kname,
            maj_min=maj_min,
            rm=rm,
            size=size,
            bsize=int(bsize),
            ro=ro,
            tp=tp,
            mountpoint=mountpoint)


@aslist
def _get_pvs_info():
    """ Collect storage info from pvs command """
    for entry in _get_cmd_output(['pvs', '--noheadings', '--separator', r':'], ':', 6):
        pv, vg, fmt, attr, psize, pfree = entry
        yield PvsEntry(
            pv=pv,
            vg=vg,
            fmt=fmt,
            attr=attr,
            psize=psize,
            pfree=pfree)


@aslist
def _get_vgs_info():
    """ Collect storage info from vgs command """
    for entry in _get_cmd_output(['vgs', '--noheadings', '--separator', r':'], ':', 7):
        vg, pv, lv, sn, attr, vsize, vfree = entry
        yield VgsEntry(
            vg=vg,
            pv=pv,
            lv=lv,
            sn=sn,
            attr=attr,
            vsize=vsize,
            vfree=vfree)


@aslist
def _get_lvdisplay_info():
    """ Collect storage info from lvdisplay command """
    for entry in _get_cmd_output(['lvdisplay', '-C', '--noheadings', '--separator', r':'], ':', 12):
        lv, vg, attr, lsize, pool, origin, data, meta, move, log, cpy_sync, convert = entry
        yield LvdisplayEntry(
            lv=lv,
            vg=vg,
            attr=attr,
            lsize=lsize,
            pool=pool,
            origin=origin,
            data=data,
            meta=meta,
            move=move,
            log=log,
            cpy_sync=cpy_sync,
            convert=convert)


@aslist
def _get_systemd_mount_info():
    """
    Collect the same storage info as provided by the systemd-mount command.

    The actual implementation no longer relies on calling the systemd-mount, but rather collects the same information
    from udev directly using pyudev. The systemd-mount output parsing has proved not to be unreliable due to
    its tabular format.
    """
    ctx = pyudev.Context()
    # Filter the devices in the same way `systemd-mount --list` does
    for device in ctx.list_devices(subsystem='block', ID_FS_USAGE='filesystem'):
        # Use 'n/a' to provide the same value for unknown output fields same way the systemd-mount does
        yield SystemdMountEntry(
            node=device.device_node,
            path=device.get('ID_PATH', default='n/a'),
            model=device.get('ID_MODEL', default='n/a'),
            wwn=device.get('ID_WWN', default='n/a'),
            fs_type=device.get('ID_FS_TYPE', default='n/a'),
            label=device.get('ID_FS_LABEL', default='n/a'),
            uuid=device.get('ID_FS_UUID', default='n/a')
        )


def get_storage_info():
    """ Collect multiple info about storage and return it """
    return StorageInfo(
        partitions=_get_partitions_info('/proc/partitions'),
        fstab=_get_fstab_info('/etc/fstab'),
        mount=_get_mount_info('/proc/mounts'),
        lsblk=_get_lsblk_info(),
        pvs=_get_pvs_info(),
        vgs=_get_vgs_info(),
        lvdisplay=_get_lvdisplay_info(),
        systemdmount=_get_systemd_mount_info())



from leapp.exceptions import RequestStopAfterPhase
from leapp.libraries.stdlib import api
from leapp.reporting import Report
from leapp.utils.report import is_inhibitor


def check():
    if [msg for msg in api.consume(Report) if is_inhibitor(msg.report)]:
        raise RequestStopAfterPhase()



from leapp.models import MultipathConfFacts, MultipathConfig, MultipathConfigOption
from leapp.reporting import Report
from leapp.snactor.fixture import current_actor_context
from leapp.utils.report import is_inhibitor


def _assert_default_checker_report(report, pathname):
    assert report['title'] == \
        'Unsupported device-mapper-multipath configuration'
    assert report['severity'] == 'high'
    assert is_inhibitor(report)
    assert pathname in report['summary']


def _assert_default_detect_report(report, pathname):
    assert report['title'] == \
        'device-mapper-multipath now defaults to detecting settings'
    assert report['severity'] == 'medium'
    assert pathname in report['summary']


def _assert_reassign_maps(report, pathname):
    assert report['title'] == \
        'device-mapper-multipath now disables reassign_maps by default'
    assert report['severity'] == 'medium'
    assert pathname in report['summary']


def test_config_all_bad(current_actor_context):
    config = MultipathConfig(
            pathname='all_bad.conf', default_path_checker='directio',
            reassign_maps=True, default_detect_checker=False,
            default_detect_prio=False, default_retain_hwhandler=False)
    facts = MultipathConfFacts(configs=[config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports and len(reports) == 3
    _assert_default_checker_report(reports[0].report, 'all_bad.conf')
    _assert_default_detect_report(reports[1].report, 'all_bad.conf')
    _assert_reassign_maps(reports[2].report, 'all_bad.conf')


def test_config_all_good(current_actor_context):
    config = MultipathConfig(
            pathname='all_good.conf', default_path_checker='tur',
            reassign_maps=False, default_detect_checker=True,
            default_detect_prio=True, default_retain_hwhandler=True)
    facts = MultipathConfFacts(configs=[config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_config_unimportant(current_actor_context):
    option = MultipathConfigOption(name='path_checker', value='rdac')
    config = MultipathConfig(
            pathname='unimportant.conf', hw_str_match_exists=True,
            ignore_new_boot_devs_exists=True, new_bindings_in_boot_exists=True,
            unpriv_sgio_exists=True, detect_path_checker_exists=True,
            overrides_hwhandler_exists=True, overrides_pg_timeout_exists=True,
            queue_if_no_path_exists=True, all_devs_section_exists=True,
            all_devs_options=[option])
    facts = MultipathConfFacts(configs=[config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_bad_then_good(current_actor_context):
    bad_config = MultipathConfig(
            pathname='all_bad.conf', default_path_checker='directio',
            reassign_maps=True, default_detect_checker=False,
            default_detect_prio=False, default_retain_hwhandler=False)
    good_config = MultipathConfig(
            pathname='all_good.conf', default_path_checker='tur',
            reassign_maps=False, default_detect_checker=True,
            default_detect_prio=True, default_retain_hwhandler=True)
    facts = MultipathConfFacts(configs=[bad_config, good_config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_good_then_bad(current_actor_context):
    good_config = MultipathConfig(
            pathname='all_good.conf', default_path_checker='tur',
            reassign_maps=False, default_detect_checker=True,
            default_detect_prio=True, default_retain_hwhandler=True)
    bad_config = MultipathConfig(
            pathname='all_bad.conf', default_path_checker='directio',
            reassign_maps=True, default_detect_checker=False,
            default_detect_prio=False, default_retain_hwhandler=False)
    facts = MultipathConfFacts(configs=[good_config, bad_config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports and len(reports) == 3
    _assert_default_checker_report(reports[0].report, 'all_bad.conf')
    _assert_default_detect_report(reports[1].report, 'all_bad.conf')
    _assert_reassign_maps(reports[2].report, 'all_bad.conf')


def test_bad_then_nothing(current_actor_context):
    bad_config = MultipathConfig(
            pathname='all_bad.conf', default_path_checker='directio',
            reassign_maps=True, default_detect_checker=False,
            default_detect_prio=False, default_retain_hwhandler=False)
    none_config = MultipathConfig(pathname='none.conf')
    facts = MultipathConfFacts(configs=[bad_config, none_config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports and len(reports) == 3
    _assert_default_checker_report(reports[0].report, 'all_bad.conf')
    _assert_default_detect_report(reports[1].report, 'all_bad.conf')
    _assert_reassign_maps(reports[2].report, 'all_bad.conf')


def test_nothing_then_bad(current_actor_context):
    bad_config = MultipathConfig(
            pathname='all_bad.conf', default_path_checker='directio',
            reassign_maps=True, default_detect_checker=False,
            default_detect_prio=False, default_retain_hwhandler=False)
    none_config = MultipathConfig(pathname='none.conf')
    facts = MultipathConfFacts(configs=[none_config, bad_config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports and len(reports) == 3
    _assert_default_checker_report(reports[0].report, 'all_bad.conf')
    _assert_default_detect_report(reports[1].report, 'all_bad.conf')
    _assert_reassign_maps(reports[2].report, 'all_bad.conf')


def test_only_bad_checker(current_actor_context):
    bad_checker_config = MultipathConfig(
            pathname='bad_checker.conf', default_path_checker='rdac',
            default_retain_hwhandler=True)
    facts = MultipathConfFacts(configs=[bad_checker_config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports and len(reports) == 1
    _assert_default_checker_report(reports[0].report, 'bad_checker.conf')


def test_only_bad_detect(current_actor_context):
    bad_detect_config = MultipathConfig(
            pathname='bad_detect.conf', default_detect_prio=True,
            default_detect_checker=False)
    facts = MultipathConfFacts(configs=[bad_detect_config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports and len(reports) == 1
    _assert_default_detect_report(reports[0].report, 'bad_detect.conf')


def test_only_bad_reassign(current_actor_context):
    bad_reassign_config = MultipathConfig(
            pathname='bad_reassign.conf', reassign_maps=True)
    facts = MultipathConfFacts(configs=[bad_reassign_config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports and len(reports) == 1
    _assert_reassign_maps(reports[0].report, 'bad_reassign.conf')


def test_different_files(current_actor_context):
    bad_detect_checker_config = MultipathConfig(
            pathname='bad_detect_checker.conf', default_detect_checker=False)
    bad_detect_prio_config = MultipathConfig(
            pathname='bad_detect_prio.conf', default_detect_prio=False)
    bad_retain_hwhandler_config = MultipathConfig(
            pathname='bad_retain_hwhandler.conf',
            default_retain_hwhandler=False)
    facts = MultipathConfFacts(
            configs=[bad_detect_checker_config, bad_detect_prio_config,
                     bad_retain_hwhandler_config])

    current_actor_context.feed(facts)
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports and len(reports) == 1
    _assert_default_detect_report(
            reports[0].report,
            'bad_detect_checker.conf, bad_detect_prio.conf and '
            'bad_retain_hwhandler.conf')



import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import opensshuseprivilegeseparationcheck
from leapp.models import OpenSshConfig, OpenSshPermitRootLogin, Report
from leapp.snactor.fixture import current_actor_context


def test_no_config(current_actor_context):
    with pytest.raises(StopActorExecutionError):
        opensshuseprivilegeseparationcheck.process(iter([]))


osprl = OpenSshPermitRootLogin(value='no')


@pytest.mark.parametrize('values,expected_report', [
    ([''], False),
    (['sandbox'], False),
    (['yes'], True),
    (['no'], True),
    (['sandbox', 'yes'], False),
    (['yes', 'sandbox'], True)])
def test_separation(current_actor_context, values, expected_report):
    for value in values:
        if value:
            current_actor_context.feed(OpenSshConfig(
                permit_root_login=[osprl],
                use_privilege_separation=value,
                deprecated_directives=[]
            ))
        else:
            current_actor_context.feed(OpenSshConfig(
                permit_root_login=[osprl],
                deprecated_directives=[]
            ))
    current_actor_context.run()
    if expected_report:
        assert current_actor_context.consume(Report)
    else:
        assert not current_actor_context.consume(Report)



import os

from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import StorageInfo, XFSPresence


def scan_xfs_fstab(data):
    mountpoints = set()
    for entry in data:
        if entry.fs_vfstype == "xfs":
            mountpoints.add(entry.fs_file)

    return mountpoints


def scan_xfs_mount(data):
    mountpoints = set()
    for entry in data:
        if entry.tp == "xfs":
            mountpoints.add(entry.mount)

    return mountpoints


def is_xfs_without_ftype(mp):
    if not os.path.ismount(mp):
        # Check if mp is actually a mountpoint
        api.current_logger().warning('{} is not mounted'.format(mp))
        return False
    try:
        xfs_info = run(['/usr/sbin/xfs_info', '{}'.format(mp)], split=True)
    except CalledProcessError as err:
        api.current_logger().warning('Error during command execution: {}'.format(err))
        return False

    for l in xfs_info['stdout']:
        if 'ftype=0' in l:
            return True

    return False


def scan_xfs():
    storage_info_msgs = api.consume(StorageInfo)
    storage_info = next(storage_info_msgs, None)

    if list(storage_info_msgs):
        api.current_logger().warning('Unexpectedly received more than one StorageInfo message.')

    fstab_data = set()
    mount_data = set()
    if storage_info:
        fstab_data = scan_xfs_fstab(storage_info.fstab)
        mount_data = scan_xfs_mount(storage_info.mount)

    mountpoints = fstab_data | mount_data
    mountpoints_ftype0 = list(filter(is_xfs_without_ftype, mountpoints))

    # By now, we only have XFS mountpoints and check whether or not it has ftype = 0
    api.produce(XFSPresence(
        present=len(mountpoints) > 0,
        without_ftype=len(mountpoints_ftype0) > 0,
        mountpoints_without_ftype=mountpoints_ftype0,
    ))



import os
import re
import shutil

from leapp.libraries.stdlib import api, CalledProcessError, run

BACKUP_DIRECTORY = '/var/lib/selinux/leapp-backup'
SEMANAGE_MODIFY_BUG = ["port", "user", "login", "fcontext", "ibpkey", "ibendport", "node", "interface"]


def list_selinux_modules():
    """
    Produce list of SELinux policy modules

    Returns list of tuples (name,priority)
    """
    try:
        semodule = run(['semodule', '-lfull'], split=True)
    except CalledProcessError:
        api.current_logger().warning('Cannot get list of selinux modules')
        return []

    modules = []
    for module in semodule.get('stdout', []):
        # Matching line such as "100 zebra             pp"
        #                       "400 virt_supplementary pp disabled"
        # "<priority> <module name> <module type - pp/cil> [disabled]"
        m = re.match(r'([0-9]+)\s+([\w-]+)\s+([\w-]+)(?:\s+([\w]+))?\s*\Z', module)
        if not m:
            # invalid output of "semodule -lfull"
            api.current_logger().warning('Invalid output of "semodule -lfull": {}'.format(module))
            continue
        modules.append((m.group(2), m.group(1)))

    return modules


# determine which (if any) udica templates where installed and install their new versions
def install_udica_templates(templates):
    if not templates:
        return

    command = ['semodule']
    for module in templates:
        command.extend(
            [
                '-X',
                str(module.priority),
                '-i',
                '/usr/share/udica/templates/{}.cil'.format(module.name)
            ]
        )

    try:
        run(command)
    except CalledProcessError as e:
        api.current_logger().warning('Error installing udica templates: {}'.format(e.stderr))


# move given file to the backup directory so that users can access it after the upgrade
def back_up_failed(module_path):
    # make sure the backup dir exists
    if not os.path.isdir(BACKUP_DIRECTORY):
        try:
            os.mkdir(BACKUP_DIRECTORY)
        except OSError:
            api.current_logger().warning('Failed to create backup directory!')
            return
    try:
        shutil.move(module_path, BACKUP_DIRECTORY)
    except OSError:
        api.current_logger().warning('Failed to back-up: {}!'.format(module_path))
        return


# Work around a "semanage import bug" by replacing "-a" (add) with -m (modify)
def modify_instead_of_add(command):
    com = command.split()
    if len(com) < 2:
        return None
    if com[0] in SEMANAGE_MODIFY_BUG and com[1] == "-a":
        com[1] = "-m"
        return " ".join(com)

    return None



import pytest

from leapp import reporting
from leapp.libraries.actor.cupscheck import check_certkey_directives
from leapp.libraries.common.testutils import create_report_mocked
from leapp.models import CupsChangedFeatures


@pytest.mark.parametrize("certkey_exists,n_reports", [(False, 0), (True, 1)])
def test_check_certkey_directives(certkey_exists, n_reports):
    facts = CupsChangedFeatures(certkey=certkey_exists)
    report_func = create_report_mocked()

    check_certkey_directives(facts, report_func)

    assert report_func.called == n_reports

    if report_func.called:
        report_fields = report_func.report_fields

        assert 'ServerKey/ServerCertificate directives' in report_fields['title']
        assert report_fields['severity'] == reporting.Severity.MEDIUM



import ipaddress
import os
import re
import shutil

from six import text_type

BackupSuffix = '.bak'

# false positives blacklist
rfp = re.compile(r'(^\s*RIPv6:::1\b)|(@\s+\[IPv6:::1\]\s+>)')

rs = re.compile(r'IPv6:[0-9a-fA-F:]*::[0-9a-fA-F:]*')


def uncompress_ipv6(ipv6):
    addr = text_type(ipv6.replace('IPv6:', ''))
    try:
        addr = 'IPv6:' + ipaddress.ip_address(addr).exploded
    except ValueError:
        addr = ipv6
    return re.sub(r':0([^:])', r':\1', re.sub(r'0+', r'0', addr))


def check_false_positives(f, l):
    return f in ['sendmail.cf', 'submit.cf'] and rfp.search(l) is not None


def sub_ipv6(m):
    return uncompress_ipv6(m.group(0))


def migrate_file(fn):
    # make backup
    shutil.copy2(fn, fn + BackupSuffix)
    with open(fn, 'w') as file_out:
        with open(fn + BackupSuffix) as file_in:
            for line in file_in:
                if rs.search(line) and not check_false_positives(os.path.basename(fn), line):
                    line = rs.sub(sub_ipv6, line)
                file_out.write(line)



from leapp.actors import Actor
from leapp.libraries.actor import blsgrubcfgonppc64
from leapp.models import DefaultGrubInfo, FirmwareFacts, GrubCfgBios, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckBlsGrubOnPpc64(Actor):
    """
    Check whether GRUB config is BLS aware on RHEL 8 ppc64le systems

    After a ppc64 system is upgraded from RHEL 8 to RHEL 9 and
    GRUB config on RHEL 8 is not yet BLS aware, the system boots
    into el8 kernel because the config is not successfully migrated by
    GRUB during the upgrade process.

    IMPORTANT NOTE: The later fix which is based on the outcome of this
    actor is applied only for virtualized ppc64le systems as we got
    unexpected behavior on bare metal ppc64le systems which needs to be
    investigated first.

    """

    name = 'check_bls_grub_onppc64'
    consumes = (DefaultGrubInfo, GrubCfgBios, FirmwareFacts)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        blsgrubcfgonppc64.process()



from six.moves.configparser import ParsingError

from leapp.libraries.common import utils
from leapp.libraries.stdlib import api, CalledProcessError, run


def read_nm_config(file_path=None):
    if file_path:
        try:
            with open(file_path, 'r') as f:
                r = f.read()
                return r
        except IOError as e:
            api.current_logger().warning('Error reading NetworkManager configuration from {}: {}'.format(file_path, e))
            return None
    else:
        try:
            # Use 'NM --print-config' to read the configurationo so
            # that the main configuration file and other files in
            # various directories get merged in the right way.
            r = run(['NetworkManager', '--print-config'], split=False)['stdout']
            return r
        except (OSError, CalledProcessError) as e:
            api.current_logger().warning('Error reading NetworkManager configuration: {}'.format(e))
            return None


def parse_nm_config(cfg):
    try:
        return utils.parse_config(cfg)
    except (ParsingError, TypeError) as e:
        api.current_logger().warning('Error parsing NetworkManager configuration: {}'.format(e))
        return None


def check_nm_dhcp(nm_cfg, parser):
    if parser.has_option('main', 'dhcp'):
        nm_cfg.dhcp = parser.get("main", "dhcp")



import os

from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import TransactionCompleted
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag


class PythonThreeTmpWorkaround(Actor):
    """
    Create the /usr/bin/python3 alternative if not exists.

    During the RPM upgrade the /usr/bin/python3 is removed because of problem
    in alternatives. The fix requires new builds of python36 on RHEL8, python3
    on RHEL 9 and alternatives on both systems. Once the internal repositories
    are updated, we can drop this. If the /usr/bin/python3 file exists,
    do nothing.
    """

    name = 'pythonthreetmpworkaround'
    consumes = (TransactionCompleted,)
    produces = ()
    tags = (RPMUpgradePhaseTag, IPUWorkflowTag)

    def process(self):
        if os.path.isfile('/usr/bin/python3'):
            api.current_logger().info('The python3 file exists. the actor can be removed probably.')
            return

        cmd = [
            'alternatives', '--install', '/usr/bin/python3', 'python3', '/usr/bin/python3.9', '1000000',
        ]

        sub_cmds = [
            ['--slave', '/usr/share/man/man1/python3.1.gz', 'python3-man', '/usr/share/man/man1/python3.9.1.gz'],
            ['--slave', '/usr/bin/pip3', 'pip3', '/usr/bin/pip3.9'],
            ['--slave', '/usr/bin/pip-3', 'pip-3', '/usr/bin/pip-3.9'],
            ['--slave', '/usr/bin/easy_install-3', 'easy_install-3', '/usr/bin/easy_install-3.9'],
            ['--slave', '/usr/bin/pydoc3', 'pydoc3', '/usr/bin/pydoc3.9'],
            ['--slave', '/usr/bin/pydoc-3', 'pydoc-3', '/usr/bin/pydoc3.9'],
            ['--slave', '/usr/bin/pyvenv-3', 'pyvenv-3', '/usr/bin/pyvenv-3.9'],
        ]

        for sub_cmd in sub_cmds:
            if os.path.exists(sub_cmd[1]):
                # some rhel 9 packages are already updated and handle
                # alternatives correctly using the --keep-foreign option
                continue
            cmd += sub_cmd

        try:
            run(cmd)
        except CalledProcessError as exc:
            raise StopActorExecutionError(
                message='Cannot create python3 alternatives; upgrade cannot be finished',
                details={'details': str(exc), 'stderr': exc.stderr},
            )



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class SELinuxFacts(Model):
    topic = SystemFactsTopic

    runtime_mode = fields.Nullable(fields.StringEnum(['enforcing', 'permissive']))
    static_mode = fields.StringEnum(['enforcing', 'permissive', 'disabled'])
    enabled = fields.Boolean()
    policy = fields.StringEnum(['targeted', 'minimum', 'mls'])
    mls_enabled = fields.Boolean()



from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import QuaggaToFrrFacts

QUAGGA_DAEMONS = [
    'babeld',
    'bgpd',
    'isisd',
    'ospf6d',
    'ospfd',
    'ripd',
    'ripngd',
    'zebra'
]


def _check_service(name, state):
    try:
        run(['systemctl', 'is-{}'.format(state), name])
        api.current_logger().debug('%s is %s', name, state)
    except CalledProcessError:
        api.current_logger().debug('%s is not %s', name, state)
        return False

    return True


def process_daemons():
    active_daemons = [daemon for daemon in QUAGGA_DAEMONS if _check_service(daemon, 'active')]
    enabled_daemons = [daemon for daemon in QUAGGA_DAEMONS if _check_service(daemon, 'enabled')]

    if active_daemons:
        api.current_logger().debug('active quaggadaemons: %s', ', '.join(active_daemons))

    if enabled_daemons:
        api.current_logger().debug('enabled quaggadaemons: %s', ', '.join(enabled_daemons))

    return QuaggaToFrrFacts(active_daemons=active_daemons, enabled_daemons=enabled_daemons)



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic, TransactionTopic


class SELinuxModule(Model):
    """SELinux module in cil including priority"""
    topic = SystemInfoTopic
    name = fields.String()
    priority = fields.Integer()
    content = fields.String()
    # lines removed due to content invalid on RHEL 8
    removed = fields.List(fields.String())


class SELinuxModules(Model):
    """
    List of selinux modules that are not part of distribution policy

    modules - list of custom policy modules (priority != 100,200)
    templates - List of installed udica templates
    """
    topic = SystemInfoTopic
    modules = fields.List(fields.Model(SELinuxModule))
    templates = fields.List(fields.Model(SELinuxModule))


class SELinuxCustom(Model):
    """SELinux customizations returned by semanage export"""
    topic = SystemInfoTopic
    commands = fields.List(fields.String())
    removed = fields.List(fields.String())


class SELinuxRequestRPMs(Model):
    """
    SELinux related RPM packages that need to be present after upgrade

    Listed packages provide types that where used in policy
    customizations (to_install), or the corresponding policy
    was installed on RHEL-7 installation with priority 200
    (to_keep).
    """
    topic = TransactionTopic
    to_keep = fields.List(fields.String(), default=[])
    to_install = fields.List(fields.String(), default=[])



from leapp.libraries.common import grub
from leapp.libraries.common.config import mock_configs
from leapp.models import GrubInfo


def _get_grub_devices_mocked():
    return ['/dev/vda', '/dev/vdb']


def test_actor_scan_grub_device(current_actor_context, monkeypatch):
    monkeypatch.setattr(grub, 'get_grub_devices', _get_grub_devices_mocked)
    current_actor_context.run(config_model=mock_configs.CONFIG)
    info = current_actor_context.consume(GrubInfo)
    assert info and info[0].orig_devices == ['/dev/vda', '/dev/vdb']
    assert len(info) == 1, 'Expected just one GrubInfo message'
    assert not info[0].orig_device_name


def test_actor_scan_grub_device_one(current_actor_context, monkeypatch):

    def _get_grub_devices_mocked():
        return ['/dev/vda']

    monkeypatch.setattr(grub, 'get_grub_devices', _get_grub_devices_mocked)
    current_actor_context.run(config_model=mock_configs.CONFIG)
    info = current_actor_context.consume(GrubInfo)
    assert info and info[0].orig_devices == ['/dev/vda']
    assert len(info) == 1, 'Expected just one GrubInfo message'
    assert info[0].orig_device_name == '/dev/vda'


def test_actor_scan_grub_device_s390x(current_actor_context, monkeypatch):
    monkeypatch.setattr(grub, 'get_grub_devices', _get_grub_devices_mocked)
    current_actor_context.run(config_model=mock_configs.CONFIG_S390X)
    assert not current_actor_context.consume(GrubInfo)



import warnings

from leapp.libraries.common.config.version import get_source_major_version

try:
    import dnf
except ImportError:
    dnf = None
    warnings.warn('Could not import the `dnf` python module.', ImportWarning)

try:
    import hawkey
except ImportError:
    hawkey = None
    warnings.warn('Could not import the `hawkey` python module.', ImportWarning)


def _create_or_get_dnf_base(base=None):
    if not base:
        # The DNF command reads /etc/yum/vars/releasever, but the DNF library does not. It parses redhat-release
        # package to retrieve system's major version which it then uses as $releasever. However, some systems might
        # have repositories only for the exact system version (including the minor number). In a case when
        # /etc/yum/vars/releasever is present, read its contents so that we can access repositores on such systems.
        conf = dnf.conf.Conf()

        # preload releasever from what we know, this will be our fallback
        conf.substitutions['releasever'] = get_source_major_version()

        # dnf on EL7 doesn't load vars from /etc/yum, so we need to help it a bit
        if get_source_major_version() == '7':
            try:
                with open('/etc/yum/vars/releasever') as releasever_file:
                    conf.substitutions['releasever'] = releasever_file.read().strip()
            except IOError:
                pass

        # load all substitutions from etc
        conf.substitutions.update_from_etc('/')

        base = dnf.Base(conf=conf)
        base.init_plugins()
        base.read_all_repos()
        # configure plugins after the repositories are loaded
        # e.g. the amazon-id plugin requires loaded repositories
        # for the proper configuration.
        base.configure_plugins()
        base.fill_sack()
    return base


def get_modules(base=None):
    """
    Return info about all module streams as a list of libdnf.module.ModulePackage objects.
    """
    if not dnf:
        return []
    base = _create_or_get_dnf_base(base)

    module_base = dnf.module.module_base.ModuleBase(base)
    # this method is absent on RHEL 7, in which case there are no modules anyway
    if not hasattr(module_base, 'get_modules'):
        return []
    return module_base.get_modules('*')[0]


def get_enabled_modules():
    """
    Return currently enabled module streams as a list of libdnf.module.ModulePackage objects.
    """
    if not dnf:
        return []

    base = _create_or_get_dnf_base()
    modules = get_modules(base)

    # if modules are not supported (RHEL 7), base.sack._moduleContainer won't exist
    # luckily in that case modules are empty and the element won't even be accessed
    return [m for m in modules if base.sack._moduleContainer.isEnabled(m)]


def map_installed_rpms_to_modules():
    """
    Map installed modular packages to the module streams they come from.
    """
    modules = get_modules()
    # empty on RHEL 7 because of no modules
    if not modules:
        return {}
    # create a reverse mapping from the RPMS to module streams
    # key: tuple of 4 strings representing a NVRA (name, version, release, arch) of an RPM
    # value: tuple of 2 strings representing a module and its stream
    rpm_streams = {}
    for module in modules:
        for rpm in module.getArtifacts():
            nevra = hawkey.split_nevra(rpm)
            rpm_key = (nevra.name, nevra.version, nevra.release, nevra.arch)
            rpm_streams[rpm_key] = (module.getName(), module.getStream())
    return rpm_streams
